"""All comparison functions."""
from processor.helper.json.json_utils import check_field_exists, get_field_value


def apply_extras(value, extras):
    """Apply any additional functionalities after evaluation."""
    for extra in extras:
        if extra == 'len':
            value = len(value) if hasattr(value, 'len') or hasattr(value, '__len__') else 0
    return value


def equality(data, loperand, roperand, is_not=False, extras=None):
    """ Compare and return value """
    value = get_field_value(data, loperand)
    eql = False
    if value:
        if extras:
            value = apply_extras(value, extras)
        if type(value) == type(roperand) and value == roperand:
            eql = True
    if is_not:
        eql = not eql
    return eql


def less_than(data, loperand, roperand, is_not=False, extras=None):
    """ Compare and return value """
    value = get_field_value(data, loperand)
    lt = False
    if value:
        if extras:
            value = apply_extras(value, extras)
        if type(value) == type(roperand) and value < roperand:
            lt = True
    if is_not:
        lt = not lt
    return lt


def less_than_equal(data, loperand, roperand, is_not=False, extras=None):
    """ Compare and return value """
    value = get_field_value(data, loperand)
    lte = False
    if value:
        if extras:
            value = apply_extras(value, extras)
        if type(value) == type(roperand) and value <= roperand:
            lte = True
    if is_not:
        lte = not lte
    return lte


def greater_than(data, loperand, roperand, is_not=False, extras=None):
    """ Compare and return value """
    value = get_field_value(data, loperand)
    gt = False
    if value:
        if extras:
            value = apply_extras(value, extras)
        if type(value) == type(roperand) and value > roperand:
            gt = True
    if is_not:
        gt = not gt
    return gt


def greater_than_equal(data, loperand, roperand, is_not=False, extras=None):
    """ Compare and return value """
    value = get_field_value(data, loperand)
    gte = False
    if value:
        if extras:
            value = apply_extras(value, extras)
        if type(value) == type(roperand) and value >= roperand:
            gte = True
    if is_not:
        gte = not gte
    return gte


def exists(data, loperand, _, is_not=False, extras=None):
    """ Compare and return value """
    present = check_field_exists(data, loperand)
    if is_not:
        present = not present
    return present



import ast
import re
import pymongo
import json
from processor.database.database import COLLECTION, get_documents
from processor.logging.log_handler import getlogger, init_logger
from processor.comparison.comparisonantlr.compare_types import EQ, NEQ, GT, GTE, LT, LTE
from processor.comparison.comparisonantlr.compare_types import compare_none, compare_int,\
    compare_float, compare_boolean, compare_str, compare_list, compare_in
from processor.helper.config.rundata_utils import get_dbtests, get_from_currentdata
from processor.helper.config.config_utils import get_test_json_dir, SINGLETEST
from processor.helper.file.file_utils import exists_file, exists_dir
from processor.helper.json.json_utils import json_from_file

IN = 'in'
comparefuncs = {
    type(None): compare_none,
    type(0): compare_int,
    type(0.1): compare_float,
    type(True): compare_boolean,
    type('abcd'): compare_str,
    type([]): compare_list,
    IN: compare_in
}
COMP = [EQ, NEQ, GT, LT, GTE, LTE]
STAR = '*'
PLUS = '+'
OPS = [STAR, PLUS]

logger = getlogger()


class RuleInterpreter:

    def __init__(self, children, **otherargs):
        self.children = children if children and isinstance(children, list) else []
        self.kwargs = otherargs
        self.lhs_operand, self.op, self.rhs_operand = self.rule_operands()
        self.snapshots = []
        logger.debug(self.kwargs)
        logger.debug(type(self.kwargs))

    def get_snapshots(self):
        return self.snapshots

    def rule_operands(self):
        lhs = self.children
        op = EQ
        rhs = ['True']
        for idx, child in enumerate(self.children):
            if child in COMP:
                lhs = self.children[:idx]
                rhs = self.children[idx+1:]
                op = child
                break
        return lhs, op, rhs


    def match_method(self, value):
        method = None
        method_args = value
        if '(' in value and ')' in value:
            method = value[:value.index('(')]
            method_args = value[value.index('(') + 1 : value.rfind(')')]
        return method, method_args


    def match_expression(self, value):
        actions = (
            (r'^(\d+)(\.\d+)?$', self.match_number),
            (r'true|false', self.match_boolean),
            (r'\'.*\'', self.match_string),
            (r'\[.*\]', self.match_array_string),
            (r'\{(\d+)\}(\[.*\].*)', self.match_array_attribute),
            (r'\{([a-zA-Z0-9_-]+)\}(\[.*\].*)', self.match_array_attribute),
            (r'\{(\d+)\}(\..*)*', self.match_attribute_array),
            (r'\{([a-zA-Z0-9_-]+)\}(\..*)*', self.match_attribute_array),
            (r'\{.*\}', self.match_dictionary_string),
            (r'.*', self.match_string),
        )
        val = None
        for regex, action in actions:
            m = re.match(regex, value, re.I)
            if m:
                logger.debug("Regex match- groups:%s, regex:%s, value: %s, function: %s ",
                             m.groups(), regex, value, action)
                val = action(value, m)
                break
        return val


    def match_number(self, value, m):
        grps = m.groups()
        if grps[1]:
            return float(value)
        else:
            return int(grps[0])


    def match_boolean(self, value, _):
        return True if value.lower() == 'true' else False


    def match_string(self, value, _):
        return value.replace("'", '')

    def match_array_string(self, value, _):
        str_val = value.replace("'", '')
        array_vals_str = str_val[1:-1]
        array_vals_str = array_vals_str.replace(' ', '')
        return array_vals_str.split(',')

    def match_dictionary_string(self, value, _):
        val = ast.literal_eval(value)
        return val

    def match_array_attribute(self, value, m):
        grps = m.groups()
        logger.info('matched grps: %s, type(grp0): %s', grps, type(grps[0]))
        doc = self.get_snaphotid_doc(grps[0])
        value = None
        if doc and grps[1]:
            value = self.get_field_value(doc, grps[1])
        return value


    def match_attribute_array(self, value, m):
        grps = m.groups()
        logger.debug('matched grps: %s, type(grp0): %s', grps, type(grps[0]))
        doc = self.get_snaphotid_doc(grps[0])
        value = None
        if doc:
            if grps[1]:
                value = self.get_field_value(doc, grps[1])
            else:
                value = doc
        return value


    def get_snaphotid_doc(self, sid):
        doc = None
        isdb_fetch = get_dbtests()
        if isdb_fetch:
            dbname = self.kwargs['dbname']
            coll = self.kwargs['snapshots'][sid] if sid in self.kwargs['snapshots'] else COLLECTION
            docs = get_documents(coll, {'snapshotId': sid}, dbname,
                                 sort=[('timestamp', pymongo.DESCENDING)], limit=1)
            logger.debug('Number of Snapshot Documents: %s', len(docs))
            if docs and len(docs):
                doc = docs[0]['json']
                snapshot = {
                    'id': docs[0]['snapshotId'],
                    'structure': docs[0]['structure'],
                    'reference': docs[0]['reference'],
                    'source': docs[0]['source'],
                    'collection': docs[0]['collection']
                }
                if 'paths' in docs[0]:
                    snapshot['paths'] = docs[0]['paths']
                else:
                    snapshot['path'] = docs[0]['path']
                self.snapshots.append(snapshot)
        else:
            json_dir = '%s%s' % (get_test_json_dir(), self.kwargs['container'])
            if exists_dir(json_dir):
                fname = '%s/snapshots/%s' % (json_dir, sid)
                if exists_file(fname):
                    json_data = json_from_file(fname)
                    if json_data and 'json' in json_data:
                        doc = json_data['json']
                        snapshot_val = {
                            'id': json_data['snapshotId'],
                            'structure': json_data['structure'],
                            'reference': json_data['reference'],
                            'source': json_data['source'],
                            'collection': json_data['collection']
                        }
                        if 'paths' in json_data:
                            snapshot_val['paths'] = json_data['paths']
                        else:
                            snapshot_val['path'] = json_data['path']
                        singletest = get_from_currentdata(SINGLETEST)
                        if singletest:
                            snapshot_val['json'] = doc
                        self.snapshots.append(snapshot_val)
        return doc


    def compare(self):
        lhs_value = self.get_value(self.lhs_operand)
        # print('LHS value: ', lhs_value)
        rhs_value = self.get_value(self.rhs_operand)
        # print('RHS value: ', rhs_value)
        logger.info('LHS: %s, OP: %s, RHS: %s', lhs_value, self.op, rhs_value)
        retval = False
        if self.op == IN:
            return compare_in(lhs_value, rhs_value, IN)
        elif type(lhs_value) == type(rhs_value):
            if type(lhs_value) in comparefuncs:
                return comparefuncs[type(lhs_value)](lhs_value, rhs_value, self.op)
        elif isinstance(lhs_value, list):
            intermediate_result = True
            for each_lhs_value in lhs_value:
                if type(each_lhs_value) == type(rhs_value):
                    if type(each_lhs_value) in comparefuncs:
                        intermediate_result = comparefuncs[type(each_lhs_value)](each_lhs_value, rhs_value, self.op)
                    else:
                        intermediate_result = False
                else:
                    intermediate_result = False
                if not intermediate_result:
                    break
            if intermediate_result:
                retval = True
        return retval


    def get_value(self, value):
        retval = None
        vals = []
        last_idx = 0
        ops = None
        for idx, child in enumerate(value):
            if child in OPS:
                vals.append((value[last_idx:idx], ops))
                last_idx = idx + 1
                ops = child
        if last_idx:
            vals.append((value[last_idx:], ops))
        if not vals:
            vals.append((value, None))
        for val in vals:
            expression_val = self.eval_expression(''.join(val[0]))
            if val[1]:
                retval = self.apply_op(val[1], retval, expression_val)
                # retval = expression_val
            else:
                retval = expression_val
        return retval

    def apply_op(self, op, val1, val2):
        retval = val2
        if op in OPS and val1 and val2 and type(val1) == type(val2):
            retval = val1 + val2
        return retval

    def is_method(self, value):
        return True if '(' in value and ')' in value else False


    def eval_expression(self, expr):
        if self.is_method(expr):
            method, method_args = self.match_method(expr)
            if method == 'exist' or method == 'exists':
                # new_args = method_args.rsplit('.', 1)[0]
                val = self.eval_expression(method_args)
            else:
                val = self.eval_expression(method_args)

            return self.apply_method(method, val, method_args)
        else:
            return self.match_expression(expr);


    def apply_method(self, method, val, method_args):
        if method == 'exist' or method == 'exists':
            # new_args = method_args.rsplit('.', 1)[-1]
            # new_args = re.sub(r'\[(.*)\]', '', new_args)
            # val = True if val and new_args in val else False
            val = True if val else False
        elif method == 'count':
            val = len(val) if hasattr(val, '__len__') else 0
        elif method == 'contain' or method == 'contains':
            self.op = 'in'
        return val

    @staticmethod
    def get_field_value(data, parameter):
        """Utility to get json value from a nested structure."""
        retval = None
        parameter = parameter[1:] if parameter.startswith('.') else parameter
        parameter = parameter[:-1] if parameter.endswith('.') else parameter
        parameter = parameter[:-2] if parameter.endswith('[]') else parameter
        # if '[]' in parameter:
        #     return None
        if data and parameter:
            fields = parameter.split('.')
            retval = data
            for field in fields:
                # print('LOOP: field: %s, retval:%s' % (field, json.dumps(retval, indent=2)))
                # print('LOOP: field: %s' % field)
                if retval:
                    if isinstance(retval, list):
                        # print("Handle Array of objects")
                        values = retval
                        retval = None
                        newvals = []
                        for val in values:
                            is_array = False
                            indexval = None
                            curval = None
                            if '[' in field and ']':
                                is_array = True
                                start = field.index('[')
                                indexval = field[start+1:-1]
                                newfield = field[:start]
                                if newfield and newfield in val and isinstance(val, dict):
                                    curval = val[newfield]
                            if is_array:
                                if isinstance(curval, list):
                                    if re.match(r'^\d+$', indexval, re.I):
                                        indexval = int(indexval)
                                        if indexval < len(curval):
                                            newvals.append(curval[indexval])
                                    elif '=' in indexval:
                                        index_fields = indexval.split('=')
                                        name = index_fields[0].replace("'", '').replace('"', '').strip()
                                        value = index_fields[-1].replace("'", '').replace('"', '').strip()
                                        for index_doc in curval:
                                            if name in index_doc and str(index_doc[name]) == value:
                                                newvals.append(index_doc)
                                                break
                                    elif '*' in indexval:
                                        # print(' #########  Handle * in array operator ###########')
                                        if curval and isinstance(curval, list):
                                            newvals.extend(curval)
                                    elif curval:
                                        newvals.append(curval[0])
                            elif field in val: # and isinstance(val, dict):
                                newvals.append(val[field])
                        if newvals:
                            retval = newvals if len(newvals) > 1 else newvals[0]
                    elif isinstance(retval, dict):
                        is_array = False
                        indexval = None
                        if '[' in field and ']':
                            is_array = True
                            start = field.index('[')
                            indexval = field[start+1:-1]
                            newfield = field[:start]
                            if newfield and newfield in retval and isinstance(retval, dict):
                                retval = retval[newfield]
                        if is_array:
                            if isinstance(retval, list):
                                if re.match(r'^\d+$', indexval, re.I):
                                    indexval = int(indexval)
                                    if indexval < len(retval):
                                        retval = retval[indexval]
                                    else:
                                        retval = None
                                elif '=' in indexval:
                                    index_fields = indexval.split('=')
                                    name = index_fields[0].replace("'", '').replace('"','').strip()
                                    value = index_fields[-1].replace("'", '').replace('"', '').strip()
                                    arrretval = None
                                    for index_doc in retval:
                                        if name in index_doc and str(index_doc[name]) == value:
                                            arrretval = index_doc
                                            break
                                    retval = arrretval
                                elif '*' in indexval:
                                    # print(' #########  Handle * in array operator ###########')
                                    # retval = retval[field]
                                    pass
                                else:
                                    retval = retval[0] if retval else None
                                    # retval = None
                            else:
                                retval = None
                        elif field in retval and isinstance(retval, dict):
                            retval = retval[field]
                        else:
                            retval = None
                    else:
                        retval = None
        return retval


# if __name__ == "__main__":
#     # data = [{
#     #         "id": 124,
#     #         "location": "eastus2",
#     #         "name": "mno-nonprod-shared-cet-eastus2-tab-as03"
#     #     },
#     #         {
#     #             "id": 125,
#     #             "location": "eastus",
#     #             "name": "mno-nonprod-shared-cet-eastus2-tab-as04"
#     #         }
#     # ]
#     # parameter = '[0].location'
#     # value = RuleInterpreter.get_field_value(data, parameter)
#     # print(value)
# #
#     otherdata = {'dbname': 'validator', 'snapshots': {}}
#     children = ["{1}[0].location", "=", "'eastus'"]
#     r_i = RuleInterpreter(children, **otherdata)
#     val = r_i.get_value(r_i.lhs_operand)
#     print(val)
#     otherdata = {'dbname': 'validator', 'snapshots': {}}
#     # children = ["{1}.location", "=", "'eastus'"]
#     children = ["{1}.location", "+", "{1}.location", "=", "'eastus'"]
#     r_i = RuleInterpreter(children, **otherdata)
#     val = r_i.get_value(r_i.lhs_operand)
#     print(val)
#     otherdata = {'dbname': 'validator', 'snapshots': {}}
#     children = ["exist", "(", "{1}.location", ")"]
#     r_i = RuleInterpreter(children, **otherdata)
#     val = r_i.eval_expression(''.join(r_i.lhs_operand))
#     otherdata = {'dbname': 'validator', 'snapshots': {}}
#     children = ["exist", "(", "{1}.location", ")"]
#     r_i = RuleInterpreter(children, **otherdata)
#     data = {'a': 1, 'b': [{'c': 'd'}, {'c': 'f'}]}
#     val = RuleInterpreter.get_field_value(data, 'b[2]')
#     print(val)

def main(argv):
    # input = FileStream(argv[1])
    from antlr4 import InputStream
    from antlr4 import CommonTokenStream
    from processor.comparison.comparisonantlr.comparatorLexer import comparatorLexer
    from processor.comparison.comparisonantlr.comparatorParser import comparatorParser
    try:
        with open(argv[1]) as f:
            for line in f:
                code = line.rstrip()
                print('#' * 75)
                print('Actual Rule: ', code)
                inputStream = InputStream(code)
                lexer = comparatorLexer(inputStream)
                stream = CommonTokenStream(lexer)
                parser = comparatorParser(stream)
                tree = parser.expression()
                print(tree.toStringTree(recog=parser))
                children = []
                for child in tree.getChildren():
                    children.append((child.getText()))
                print("All the parsed tokens: ", children)
                if len(argv) == 2:
                    otherdata = {'dbname': 'validator', 'container': 'container1', 'snapshots': {}}
                    r_i = RuleInterpreter(children, **otherdata)
                    lval = r_i.get_value(r_i.lhs_operand)
                    print("LHS: {}".format(json.dumps(lval) if lval else "None"))
                    rval = r_i.get_value(r_i.rhs_operand)
                    print("RHS: {}".format(rval))
                    result = r_i.compare()
                    print("RESULT: {}".format("PASS" if result else "FAIL"))
                print('*' * 50)
        return True
    except Exception as ex:
        print("Exception: %s" % ex)
        return False


if __name__ == '__main__':
    import sys
    from processor.logging.log_handler import getlogger, init_logger, NONE
    init_logger(NONE)
    main(sys.argv)



import time
import os
import hashlib
import pymongo
from processor.helper.config.config_utils import config_value
from processor.helper.config.rundata_utils import get_dbtests
from processor.logging.log_handler import getlogger
from processor.database.database import insert_one_document, COLLECTION, DATABASE, DBNAME, \
    get_collection_size, create_indexes
from processor.helper.json.json_utils import get_field_value, json_from_file, store_snapshot, make_snapshots_dir
from processor.helper.file.file_utils import exists_file, exists_dir

logger = getlogger()

class TemplateProcessor:
    """
    Base Template Processor for process template 
    parameters:
        node : node structure from snapshot JSON
        container: name of container on which the testcase is running
        dbname: collection name in which the data will store
        snapshot_source: name of the connector file
        snapshot_data: stores the snapshot results, require for run the test
        repopath: path to the temporary directory where the repository files are available
        processed_template: require to update this field with the processed template json
        paths: list of paths from the node structure json
        dir_path: path to the directory where the repository files are available, if `folderPath` is present in node.
                  otherwise it will same as `repopath`
        snapshot: snapshot structure
    """

    def __init__(self, node, **kwargs):
        self.node = node
        self.container = kwargs.get("container")
        self.dbname = kwargs.get("dbname")
        self.snapshot_source = kwargs.get("snapshot_source")
        self.connector_data = kwargs.get("connector_data")
        self.snapshot_data = kwargs.get("snapshot_data")
        self.repopath = kwargs.get("repopath")
        self.snapshot = kwargs.get("snapshot")
        self.processed_template = {}
        self.paths = []
        self.dir_path = ""

    def create_database_record(self):
        """
        returns a generated snapshot object from processed template for store it in database or file system
        """
        collection = self.node['collection'] if 'collection' in self.node else COLLECTION
        parts = self.snapshot_source.split('.')

        ref = ""
        if "branchName" in self.connector_data:
            ref = self.connector_data["branchName"]
        elif "folderPath" in self.connector_data:
            ref = self.connector_data["folderPath"]

        db_record = {
            "structure": self.connector_data["type"],
            "error": self.processed_template['error'] if 'error' in self.processed_template else None,
            "reference": ref,
            "source": parts[0],
            "paths": self.node["paths"],
            "timestamp": int(time.time() * 1000),
            "queryuser": self.connector_data['username'] if 'username' in self.connector_data else None,
            "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
            "node": self.node,
            "snapshotId": self.node['snapshotId'],
            "collection": collection.replace('.', '').lower(),
            "json": self.processed_template
        }
        return db_record
    
    def store_data_record(self):
        """
        creates the indexes on collection and stores the data record in database or creates
        the generated snapshot at file system
        """
        data_record = self.create_database_record()
        if get_dbtests():
            if get_collection_size(self.node['collection']) == 0:
                #Creating indexes for collection
                create_indexes(
                    self.node['collection'], 
                    config_value(DATABASE, DBNAME), 
                    [
                        ('snapshotId', pymongo.ASCENDING),
                        ('timestamp', pymongo.DESCENDING)
                    ]
                )

                create_indexes(
                    self.node['collection'], 
                    config_value(DATABASE, DBNAME), 
                    [
                        ('_id', pymongo.DESCENDING),
                        ('timestamp', pymongo.DESCENDING),
                        ('snapshotId', pymongo.ASCENDING)
                    ]
                )
            insert_one_document(data_record, self.node['collection'], self.dbname, check_keys=False)
        else:
            snapshot_dir = make_snapshots_dir(self.container)
            if snapshot_dir:
                store_snapshot(snapshot_dir, data_record)
            
        if 'masterSnapshotId' in self.node:
            self.snapshot_data[self.node['snapshotId']] = self.node['masterSnapshotId']
        else:
            self.snapshot_data[self.node['snapshotId']] = False if ('error' in data_record and data_record['error']) else True
        
        self.node['status'] = 'active'

    def process_template(self, paths):
        """
        process the files stored at specified paths and returns the processed template. 
        It should be override by child class for process the template files.
        """
        return {}
    
    def is_template_file(self, file_path):
        """
        checks that file located at given path is template file or not and return boolean value.
        It should be override by child class.
        """
        return False
    
    def is_parameter_file(self, file_path):
        """
        check that file located at given path is parameter file or not and return boolean value. 
        It should be override by child class.
        """
        return False

    def populate_template_snapshot(self):
        """
        process the snapshot and returns the updated `snapshot_data` which is require for run the test
        """
        self.dir_path = get_field_value(self.connector_data, 'folderPath')
        if not self.dir_path:
            self.dir_path = self.repopath 

        self.paths = get_field_value(self.node, 'paths')
        if not self.paths or not isinstance(self.paths, list):
            self.node['status'] = 'inactive'
            logger.error("Invalid json : `paths` field is missing in node or it is not a list")
            return self.snapshot_data

        self.processed_template = self.process_template(self.paths)
        if self.processed_template:
            self.store_data_record()
            self.node['status'] = 'active'
        else:
            self.node['status'] = 'inactive'
        return self.snapshot_data
    
    def create_node_record(self, generated_template_file_list, count):
        """
        It is used in crawler. It creates a list of node objects for each generated snapshots.
        """
        nodes = []
        collection = get_field_value(self.node, 'collection')
        test_user = get_field_value(self.snapshot, 'testUser')
        source = get_field_value(self.snapshot, 'source')
        master_snapshot_id = get_field_value(self.node, 'masterSnapshotId')
        
        for template_node in generated_template_file_list:
            count = count + 1
            node_dict = {
                "snapshotId": '%s%s' % (master_snapshot_id, str(count)),
                "type": self.node["type"],
                "collection": collection,
                "paths": template_node["paths"],
                "status": template_node['status'],
                "validate": template_node['validate']
            }
            nodes.append(node_dict)
        return nodes, count

    def populate_sub_directory_snapshot(self, file_path, base_dir_path, sub_dir_path, snapshot, dbname, node, snapshot_data, count):
        """
        Iterate the subdirectories for process each files inside the directory.
        """
        logger.info("Finding files in : %s" % sub_dir_path)
        dir_path = str('%s/%s' % (base_dir_path, sub_dir_path)).replace('//', '/')

        template_file_list = []
        parameter_file_list = []

        if exists_dir(dir_path):
            list_of_file = os.listdir(dir_path)
            for entry in list_of_file:
                new_dir_path = ('%s/%s' % (dir_path, entry)).replace('//', '/')
                new_sub_directory_path = ('%s/%s' % (sub_dir_path, entry)).replace('//', '/')
                if exists_dir(new_dir_path):
                    count = self.populate_sub_directory_snapshot(file_path, base_dir_path, new_sub_directory_path, snapshot, dbname, node, snapshot_data, count)
                elif exists_file(new_dir_path):
                    if self.is_parameter_file(new_dir_path):
                        parameter_file_list.append(new_sub_directory_path)
                    elif self.is_template_file(new_dir_path):
                        template_file_list.append(new_sub_directory_path)
            
            generated_template_file_list = []
            if template_file_list:
                for template_file in template_file_list:
                    template_file_path = str('%s/%s' % (base_dir_path, template_file)).replace('//', '/')
                    if parameter_file_list:
                        for parameter_file in parameter_file_list:
                            parameter_file_json_path = str('%s/%s' % (base_dir_path, parameter_file)).replace('//', '/')
                            paths = [
                                template_file,
                                parameter_file
                            ]
                            template_json = self.process_template(paths)
                            if template_json:
                                generated_template_file_list.append({
                                    "paths" : [
                                        ("%s/%s" % (file_path, template_file)).replace("//", "/"),
                                        ("%s/%s" % (file_path, parameter_file)).replace("//", "/")
                                    ],
                                    "status" : "active",
                                    "validate" : self.node['validate'] if 'validate' in self.node else True
                                })
                            else:
                                generated_template_file_list.append({
                                    "paths" : [
                                        ("%s/%s" % (file_path, template_file)).replace("//", "/"),
                                        ("%s/%s" % (file_path, parameter_file)).replace("//", "/")
                                    ],
                                    "status" : "inactive",
                                    "validate" : self.node['validate'] if 'validate' in self.node else True
                                })
                    else:
                        paths = [
                            template_file
                        ]
                        template_json = self.process_template(paths)
                        if template_json:
                            generated_template_file_list.append({
                                "paths" : [
                                    ("%s/%s" % (file_path, template_file)).replace("//", "/")
                                ],
                                "status" : "active",
                                "validate" : node['validate'] if 'validate' in node else True
                            })
                        else:
                            generated_template_file_list.append({
                                "paths" : [
                                    ("%s/%s" % (file_path, template_file)).replace("//", "/")
                                ],
                                "status" : "inactive",
                                "validate" : node['validate'] if 'validate' in node else True
                            })
                    
                nodes, count = self.create_node_record(generated_template_file_list, count)
                if self.node['masterSnapshotId'] not in self.snapshot_data or not isinstance(self.snapshot_data[self.node['masterSnapshotId']], list):
                    self.snapshot_data[self.node['masterSnapshotId']] = []
                self.snapshot_data[self.node['masterSnapshotId']] = self.snapshot_data[self.node['masterSnapshotId']] + nodes
        return count

    def populate_all_template_snapshot(self):
        """
        crawler function for populate all the files located at given paths and generate returns the updated `snapshot_data`
        """
        root_dir_path = get_field_value(self.connector_data, 'folderPath')
        if not root_dir_path:
            root_dir_path = self.repopath 

        self.paths = get_field_value(self.node, 'paths')
        if self.paths and isinstance(self.paths, list):
            count = 0
            for path in self.paths:
                self.dir_path = str('%s/%s' % (root_dir_path, path)).replace('//', '/')
                if exists_dir(self.dir_path):
                    list_of_file = os.listdir(self.dir_path)
                    for entry in list_of_file:
                        count = self.populate_sub_directory_snapshot(path, self.dir_path, entry, self.snapshot, self.dbname, self.node, self.snapshot_data, count)
                else:
                    logger.error("Invalid path : directory does not exist : " + self.dir_path)
        else:
            logger.error("Invalid json : `paths` field is missing for 'arm' node type or it is not a list")    
        
        return self.snapshot_data



from processor.helper.json.json_utils import get_field_value_with_default, get_field_value, json_from_file
from processor.helper.file.file_utils import exists_file, exists_dir
from processor.helper.httpapi.restapi_azure import json_source
from processor.connector.vault import get_vault_data
from processor.connector.snapshot_custom import get_custom_data, git_clone_dir
from processor.logging.log_handler import getlogger
from subprocess import Popen, PIPE
import tempfile
import re
import os
import urllib

logger = getlogger()

def run_subprocess_cmd(cmd, ignoreerror=False, maskoutput=False, outputmask="Error output is masked"):
    """ Run a sub-process command"""
    result = ''
    error_result = None
    if cmd:
        if isinstance(cmd, list):
            cmd = ' '.join(cmd)
        myprocess = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        out, err = myprocess.communicate()
        result = out.rstrip()
        error_result = err.rstrip() if err else None
        if isinstance(result, bytes):
            result = result.decode()
        if error_result and isinstance(error_result, bytes):
            error_result = error_result.decode()
        if not ignoreerror and error_result:
            if maskoutput:
                logger.info("OUTPUT: %s, ERROR: %s", outputmask, outputmask)
            else:
                logger.info("CMD: '%s', OUTPUT: %s, ERROR: %s", cmd, result, error_result)
    return error_result, result

def validate_snapshot_data(snapshot_json, document_json):
    validate = True
    if "snapshots" not in snapshot_json:
        logger.info("Invalid json: 'Snapshots' field does not exists.")
        return False

    snapshots = snapshot_json["snapshots"]
    if not isinstance(snapshots, list):
        logger.info("Invalid json: 'Snapshots' field is not type list.")
        return False

    # Add validations based on snapshot type
    if validate:
        document_json["snapshots"] = snapshot_json["snapshots"]
    
    return validate

def validate_master_snapshot_data(master_snapshot_json, document_json):
    validate = True
    
    connector_users = get_field_value(document_json, "connectorUsers")
    if not connector_users:
        logger.info("Invalid snapshot: 'connectorUsers' field does not exists or it is empty.")
        return False

    if "snapshots" not in master_snapshot_json:
        logger.info("Invalid json: 'Snapshots' field does not exists.")
        return False

    snapshots = master_snapshot_json["snapshots"]
    if not isinstance(snapshots, list):
        logger.info("Invalid json: 'Snapshots' field is not type list.")
        return False

    for snapshot in snapshots:
        if "type" not in snapshot:
            logger.info("Invalid json: 'type' field is not exists in snapshot.")
            validate = False
            break

        if "connectorUser" not in snapshot:
            logger.info("Invalid json: 'connectorUser' field is not exists in snapshot.")
            validate = False
            break
        else:
            found_connector_user = False
            for connector_user in connector_users:
                if connector_user["id"] == snapshot["connectorUser"]:
                    logger.info(connector_user.items())
                    found_connector_user = True
                    for key, value in connector_user.items():
                        if key != "id":
                            snapshot[key] = value
        
            if not found_connector_user:
                logger.info("Invalid json: given 'connectorUser' not found in 'connectorUsers'.")
                validate = False
                break
            
        if "nodes" not in snapshot:
            logger.info("Invalid json: 'nodes' field is not exists in snapshot.")
            validate = False
            break

        if not isinstance(snapshot["nodes"], list):
            logger.info("Invalid json: 'snapshots -> nodes' field is not type list.")
            validate = False
            break

        for node in snapshot["nodes"]:
            if "masterSnapshotId" not in node:
                logger.info("Invalid json: 'masterSnapshotId' field is not exists in node.")
                validate = False
                break

            if snapshot["type"] == "aws" and "arn" not in node:
                logger.info("Invalid json: 'arn' field is not exists in node.")
                validate = False
                break

            elif snapshot["type"] != "aws" and "type" not in node:
                logger.info("Invalid json: 'type' field is not exists in node.")
                validate = False
                break

            if "collection" not in node:
                logger.info("Invalid json: 'collection' field is not exists in node.")
                validate = False
                break
        
        if not validate:
            break

    if validate:
        document_json.pop("connector")
        document_json.pop("remoteFile")
        document_json.pop("connectorUsers")
        document_json["snapshots"] = master_snapshot_json["snapshots"]
    
    return validate    

def validate_test_data(test_json, document_json):
    validate = True
    
    if "testSet" not in test_json:
        logger.info("Invalid json: 'testSet' field does not exists.")
        return False

    testsets = test_json["testSet"]
    if not isinstance(testsets, list):
        logger.info("Invalid json: 'testSet' field is not type list.")
        return False

    for testset in testsets:
        if "testName" not in testset:
            logger.info("Invalid json: 'testName' field is not exists in testset.")
            validate = False
            break
            
        if "cases" not in testset:
            logger.info("Invalid json: 'cases' field is not exists in testset.")
            validate = False
            break

        if not isinstance(testset["cases"], list):
            logger.info("Invalid json: 'testset -> cases' field is not type list.")
            validate = False
            break

        for case in testset["cases"]:
            if "testId" not in case:
                logger.info("Invalid json: 'testId' field is not exists in case.")
                validate = False
                break
        
        if not validate:
            break

    if validate:
        document_json.pop("connector")
        document_json.pop("remoteFile")
        document_json["testSet"] = test_json["testSet"]
    
    return validate

def validate_master_test_data(master_test_json, document_json):
    validate = True
    
    if "testSet" not in master_test_json:
        logger.info("Invalid json: 'testSet' field does not exists.")
        return False

    testsets = master_test_json["testSet"]
    if not isinstance(testsets, list):
        logger.info("Invalid json: 'testSet' field is not type list.")
        return False

    for testset in testsets:
        if "masterTestName" not in testset:
            logger.info("Invalid json: 'masterTestName' field is not exists in testset.")
            validate = False
            break
            
        if "cases" not in testset:
            logger.info("Invalid json: 'cases' field is not exists in testset.")
            validate = False
            break

        if not isinstance(testset["cases"], list):
            logger.info("Invalid json: 'testset -> cases' field is not type list.")
            validate = False
            break

        for case in testset["cases"]:
            if "masterTestId" not in case:
                logger.info("Invalid json: 'masterTestId' field is not exists in case.")
                validate = False
                break
        
        if not validate:
            break

    if validate:
        document_json.pop("connector")
        document_json.pop("remoteFile")
        document_json["testSet"] = master_test_json["testSet"]
    
    return validate

def pull_json_data(document_json):
    """
    Pull the JSON data from the git based on filetype and it will update the document json.
    """
    dirpath = None
    connector = get_field_value(document_json, "connector")
    file_location = get_field_value(document_json, "remoteFile")
    file_type = get_field_value(document_json, "fileType")

    if not connector:
        logger.info("Invalid snapshot: 'connector' field does not exists or it is empty.")
        return dirpath, False

    if not file_location:
        logger.info("Invalid snapshot: 'remoteFile' field does not exists or it is empty.")
        return dirpath, False

    if not file_type:
        logger.info("Invalid snapshot: 'fileType' field does not exists or it is empty.")
        return dirpath, False

    connector_data = get_custom_data(connector)
    baserepo, repopath = git_clone_dir(connector_data)
    
    if repopath:
        json_path = '%s/%s' % (repopath, file_location)
        dirpath = os.path.dirname(json_path)
        file_path = json_path.replace('//', '/')
        json_data = json_from_file(file_path, escape_chars=['$'])

        validate = False
        if json_data:
            if file_type == "snapshot":
                validate = validate_snapshot_data(json_data, document_json)

            if file_type == "masterSnapshot":
                validate = validate_master_snapshot_data(json_data, document_json)

            if file_type == "test":
                validate = validate_test_data(json_data, document_json)

            if file_type == "mastertest":
                validate = validate_master_test_data(json_data, document_json)
        
        return dirpath, validate
    else:
        raise Exception('Require valid fields for populate JSON are not present!')

    return dirpath, False


"""
snapshot connector file for aws type snapshot. When the 'type' is 'aws', the
resource objects have to be fetched from the AWS interface.
The connection could use python SDK libraries or https calls or other known methods
provided by the provider.
Here in case of AWS, have used boto python SDK released and maintained by AWS.
boto library provides different 'describe_' functions and using this functionality,
the snapshot mentions which describe function has to be called, so 'type' could
be 'security_groups' , 'instances', 'regions' and using python reflection capabilities
a callable from describe_security_groups gets callable for this function to return security groups.
The describe_ functions lots of methods to query AWS resources, since we know what resource we need
we always query by id, so AWS snapshots shall always have {"id": "123de23"} to uniquely
identify the resource object.
"""
import json
import hashlib
import time
import copy
import pymongo
import os
from boto3 import client
from boto3 import Session
from processor.helper.file.file_utils import exists_file
from processor.logging.log_handler import getlogger
from processor.helper.config.rundata_utils import put_in_currentdata, get_dbtests, get_from_currentdata
from processor.helper.json.json_utils import get_field_value, json_from_file,\
    collectiontypes, STRUCTURE, make_snapshots_dir, store_snapshot
from processor.connector.vault import get_vault_data
from processor.helper.config.config_utils import config_value, get_test_json_dir, CUSTOMER
from processor.database.database import insert_one_document, sort_field, get_documents,\
    COLLECTION, DATABASE, DBNAME, get_collection_size, create_indexes
from processor.helper.httpapi.restapi_azure import json_source
from processor.helper.httpapi.restapi_azure import get_client_secret
from processor.connector.snapshot_utils import validate_snapshot_nodes
from processor.connector.arn_parser import arnparse

logger = getlogger()
_valid_service_names = Session().get_available_services()


def _validate_client_name(client_name):
    """
    A private function to validate whether a given client provided
    in snapshot or aws connector is a valid service in Boto3
    """
    return client_name is not None and client_name.lower() in _valid_service_names


def get_aws_data(snapshot_source):
    """
    The AWS source object to be fetched from database or the filesystem
    The initial configuration for database is 'validator' and collection
    is 'structures', whereas for the filesystem the path to fetch the
    'structures' is  $SOLUTIONDIR/realm/<structure>.json
    """
    sub_data = {}
    if json_source():
        dbname = config_value(DATABASE, DBNAME)
        collection = config_value(DATABASE, collectiontypes[STRUCTURE])
        parts = snapshot_source.split('.')
        qry = {'name': parts[0]}
        sort = [sort_field('timestamp', False)]
        docs = get_documents(collection, dbname=dbname, sort=sort, query=qry, limit=1)
        logger.info('Number of AWS structure Documents: %d', len(docs))
        if docs and len(docs):
            sub_data = docs[0]['json']
    else:
        json_test_dir = get_test_json_dir()
        file_name = '%s.json' % snapshot_source if snapshot_source and not \
            snapshot_source.endswith('.json') else snapshot_source
        aws_source = '%s/../%s' % (json_test_dir, file_name)
        logger.info('AWS source: %s', aws_source)
        if exists_file(aws_source):
            sub_data = json_from_file(aws_source)
    return sub_data


def _get_aws_function(awsclient, node):
    """ 
    A private function to get the function which has to be called by the
    boto3 client object to get snapshot data.
    """
    describe_function_str = get_aws_describe_function(node)
    if describe_function_str:
        describe_function = getattr(awsclient, describe_function_str, None)
        if describe_function and callable(describe_function):
            return describe_function
    
    function_str = _get_callable_method_from_node(node)
    if function_str:
        callable_function = getattr(awsclient, function_str, None)
        if callable_function and callable(callable_function):
            return callable_function

    
def _get_callable_method_from_node(node):
    """Callable Method from node using python reflection mechanism"""
    _fn_str = None
    if node and 'type' in node and node['type']:
        _fn_str = node['type']
    return _fn_str
    

def get_aws_describe_function(node):
    """Describe function for the node using python reflection mechanism"""
    describe_fn_str = None
    if node and 'type' in node and node['type']:
        describe_fn_str = 'describe_%s' % node['type']
    return describe_fn_str


def get_node(awsclient, node, snapshot_source):
    """
    Fetch node from aws using connection. In this case using boto API's
    describe functions.
    """

    collection = node['collection'] if 'collection' in node else COLLECTION
    parts = snapshot_source.split('.')
    function_to_call = None
    db_record = {
        "structure": "aws",
        "error": None,
        "reference": "",
        "source": parts[0],
        "path": '',
        "timestamp": int(time.time() * 1000),
        "queryuser": "",
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": node['snapshotId'],
        "collection": collection.replace('.', '').lower(),
        "json": {}  # Refactor when node is absent it should None, when empty object put it as {}
    }
    detail_methods = get_field_value(node, "detailMethods")
    if detail_methods is None:
        function_to_call = _get_aws_function(awsclient, node)
        if function_to_call and callable(function_to_call):
            queryval = get_field_value(node, 'id')
            try:
                data = function_to_call(**queryval)
                if data:
                    db_record['json'] = data
                    checksum = get_checksum(data)
                    if checksum:
                        db_record['checksum'] = checksum
                    else:
                        put_in_currentdata('errors', data)
                        logger.info("Describe function does not exist: %s", str(function_to_call))
                        db_record['error'] = "Describe function does not exist: %s" % str(function_to_call)
            except Exception as ex:
                logger.info('Describe function exception: %s', ex)
                db_record['error'] = 'Describe function exception: %s' % ex
        else:
            logger.info('Invalid function exception: %s', str(function_to_call))
            db_record['error'] = 'Invalid function exception: %s' % str(function_to_call)
    else:
        json_to_put = {}
        arn_str = get_field_value(node, "arn")
        db_record["path"] = arn_str if arn_str else ""
        
        arn_obj = arnparse(arn_str)
        client_str = arn_obj.service
        resourceid = arn_obj.resource
        for each_method_str in detail_methods:
            function_to_call = getattr(awsclient, each_method_str, None)
            if function_to_call and callable(function_to_call):
                params = _get_function_kwargs(arn_str, each_method_str, json_to_put)
                try:
                    data = function_to_call(**params)
                    if data:
                        json_to_put.update(data) 
                except Exception as ex:
                    logger.info('Describe function exception: %s', ex)
                    db_record['error'] = 'Describe function exception: %s' % ex
            else:
                logger.info('Invalid function exception: %s', str(function_to_call))
                db_record['error'] = 'Invalid function exception: %s' % str(function_to_call)
        db_record['json'] = json_to_put
    return db_record


def _get_resources_from_list_function(response, method):
    """
    Fetches the resources id from different responses
    and returns a list of responses.
    """
    if method == 'list_buckets':
        return [x['Name'] for x in response['Buckets']]
    elif method == 'describe_instances':
        final_list = []
        for reservation in response['Reservations']:
            for instance in reservation['Instances']:
                final_list.append(instance['InstanceId'])
        return final_list
    elif method == 'describe_db_instances':
        return [x['DBInstanceIdentifier'] for x in response['DBInstances']]
    elif method == 'describe_load_balancers':
        return [x['LoadBalancerName'] for x in response['LoadBalancerDescriptions']]
    elif method == 'list_certificates':
        return [x['CertificateArn'] for x in response['CertificateSummaryList']]        
    elif method == 'list_stacks':
        return [x['StackName'] for x in response['StackSummaries']]        
    elif method == 'list_trails':
        return [x['Name'] for x in response['Trails']]        
    elif method in ['describe_stacks', 'list_trails']:
        return [x['StackName'] for x in response['Stacks']]        
    elif method == 'get_rest_apis':
        return [x['id'] for x in response['items']]   
    elif method == 'list_users':
        return [x['UserName'] for x in response['Users']]
    elif method == 'list_roles':
        return [x['RoleName'] for x in response['Roles']]     
    elif method == 'list_hosted_zones':
        return [x['Id'] for x in response['HostedZones']]
    elif method == 'list_keys':
        return [x.get('KeyId') for x in response['Keys']]
    elif method == 'list_tables':
        return response.get("TableNames")
    elif method == 'list_backups':
        return [x.get('BackupArn',"") for x in response['BackupSummaries']]
        return response.get("TableNames")
    elif method == 'list_task_definitions':
        return response.get('taskDefinitionArns')
    elif method == 'list_clusters':
        return response.get("clusters")
    elif method == 'describe_replication_groups':
        return [x.get('ReplicationGroupId') for x in response['ReplicationGroups']]
    elif method == 'list_streams':
        return response.get("StreamNames")
    elif method == 'list_functions':
        return [x.get('FunctionName',"") for x in response['Functions']]
    elif method == 'describe_clusters':
        return [x.get('ClusterIdentifier',"") for x in response['Clusters']]
    elif method == 'list_topics':
        return [x.get('TopicArn',"") for x in response['Topics']]
    elif method == 'list_queues':
        return response.get("QueueUrls")  
    elif method == 'list_domain_names':
        return [x.get('DomainName') for x in response['DomainNames']] 
    elif method == 'describe_configuration_recorders':
        return [x.get('name') for x in response['ConfigurationRecorders']] 
    elif method == 'list_distributions':
        return [x.get('Id') for x in response['DistributionList']['Items']]
    elif method == 'describe_vpn_gateways':
        return [x.get('VpnGatewayId') for x in response['VpnGateways']]
    elif method == 'describe_file_systems':
        return [x.get('FileSystemId') for x in response['FileSystems']]
    elif method == 'describe_parameters':
        return [x.get('Name') for x in response['Parameters']]
    elif method == 'describe_cache_subnet_groups':
        return [x.get('CacheSubnetGroupName') for x in response['CacheSubnetGroups']]
    elif method == 'describe_route_tables':
        return [x.get('RouteTableId') for x in response['RouteTables']]
    elif method == 'describe_network_acls':
        return [x.get('NetworkAclId') for x in response['NetworkAcls']]
    elif method == 'describe_event_subscriptions':
        return [x.get('EventSubscriptionArn').split(':')[-1] for x in response['EventSubscriptionsList']]
    elif method == 'describe_db_snapshots':
        return [x.get('DBSnapshotIdentifier') for x in response['DBSnapshots']]
    else:
        return []

def get_all_nodes(awsclient, node, snapshot, connector):
    """ Fetch all the nodes from the cloned git repository in the given path."""
    db_records = []
    arn_string = "arn:aws:%s:%s::%s"
    collection = node['collection'] if 'collection' in node else COLLECTION
    snapshot_source = get_field_value(snapshot, 'source')
    parts = snapshot_source.split('.')
    d_record = {
        "structure": "aws",
        "reference": "",
        "source": parts[0],
        "path": '',
        "timestamp": int(time.time() * 1000),
        "queryuser": "",
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": None,
        "masterSnapshotId": node['masterSnapshotId'],
        "collection": collection.replace('.', '').lower(),
        "json": {}
    }
    list_function_name = get_field_value(node, 'listMethod')
    if list_function_name:
        list_function = getattr(awsclient, list_function_name, None)
        if list_function and callable(list_function):
            try:
                list_kwargs = _get_list_function_kwargs(awsclient.meta._service_model.service_name, list_function_name)
                response = list_function(**list_kwargs)
                list_of_resources = _get_resources_from_list_function(response, list_function_name)
            except Exception as ex:
                list_of_resources = []
            detail_methods = get_field_value(node, 'detailMethods')
            for each_resource in list_of_resources:
                type_list = []
                if "arn:" in each_resource:
                    resource_arn = each_resource
                else:
                    resource_arn = arn_string %(awsclient.meta._service_model.service_name,
                        awsclient.meta.region_name, each_resource)
                for each_method_str in detail_methods:
                    each_method = getattr(awsclient, each_method_str, None)
                    if each_method and callable(each_method):
                        type_list.append(each_method_str)
                db_record = copy.deepcopy(d_record)
                db_record['detailMethods'] = type_list
                db_record['arn'] = resource_arn
                db_records.append(db_record)

    return db_records

def get_checksum(data):
    """ Get the checksum for the AWS data fetched."""
    checksum = None
    try:
        data_str = json.dumps(data, default=str)
        checksum = hashlib.md5(data_str.encode('utf-8')).hexdigest()
    except:
        pass
    return checksum

def _get_list_function_kwargs(service, function_name):
    if service == "cloudformation" and function_name == 'list_stacks':
        return {
            'StackStatusFilter' : ['CREATE_IN_PROGRESS', 'CREATE_COMPLETE', 'ROLLBACK_IN_PROGRESS',\
            'ROLLBACK_COMPLETE', 'UPDATE_IN_PROGRESS', 'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS', \
            'UPDATE_COMPLETE', 'UPDATE_ROLLBACK_IN_PROGRESS', 'UPDATE_ROLLBACK_FAILED', \
            'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS', 'UPDATE_ROLLBACK_COMPLETE', \
            'REVIEW_IN_PROGRESS', 'IMPORT_IN_PROGRESS', 'IMPORT_COMPLETE', 'IMPORT_ROLLBACK_IN_PROGRESS', \
            'IMPORT_ROLLBACK_FAILED', 'IMPORT_ROLLBACK_COMPLETE']
        }
    else:
        return {}

def _get_function_kwargs(arn_str, function_name, existing_json):
    """Fetches the correct keyword arguments for different detail functions"""
    arn = arnparse(arn_str)
    client_str = arn.service
    resource_id = arn.resource
    if client_str == "s3":
        return {'Bucket' : resource_id}
    elif client_str == "rds" and function_name in ["describe_db_instances",\
        "describe_db_snapshots"]:
        return {
            'DBInstanceIdentifier': resource_id
        }
    elif client_str == "ec2" and function_name == "describe_instance_attribute":
        return {
            'Attribute': 'instanceType',
            'InstanceId': resource_id
        }
    elif client_str == "ec2" and function_name in ["describe_instances", "monitor_instances"]:
        return {
            'InstanceIds': [resource_id]
        }
    elif client_str == "ec2" and function_name == "describe_images":
        try:
            imageid = existing_json['Reservations'][0]['Instances'][0]['ImageId']
        except:
            imageid = ""
        return {
            'ImageIds': [imageid]
        }
    elif client_str == "ec2" and function_name == "describe_volumes":
        try:
            volumeid = existing_json['Reservations'][0]['Instances'][0]['BlockDeviceMappings'][0]['Ebs']['VolumeId']
        except:
            volumeid = ""
        return {
            'VolumeIds': [volumeid]
        }
    elif client_str == "ec2" and function_name == "describe_security_groups":
        try:
            groups = existing_json['Reservations'][0]['Instances'][0]['SecurityGroups']
            groupsidlist = [x['GroupId'] for x in groups]
        except:
            volumeid = []
        return {
            'GroupIds': groupsidlist
        }
    elif client_str == "ec2" and function_name == "describe_vpcs":
        try:
            vpicid = existing_json['Reservations'][0]['Instances'][0]['VpcId']
        except:
            vpicid = ""
        return {
            'VpcIds': [vpicid]
        }
    elif client_str == "ec2" and function_name == "describe_subnets":
        try:
            subnetid = existing_json['Reservations'][0]['Instances'][0]['SubnetId']
        except:
            subnetid = ""
        return {
            'SubnetIds': [subnetid]
        }
    elif client_str == "ec2" and function_name == "describe_snapshots":
        try:
            ownerid = existing_json['Reservations'][0]['OwnerId']
        except:
            ownerid = ""
        return {
            'OwnerIds': [ownerid]
        }
    elif client_str == "ec2" and function_name == "describe_snapshot_attribute":
        try:
            snapshot_id = existing_json['Snapshots'][0]['SnapshotId']
        except:
            snapshot_id = ""
        return {
            'SnapshotId': snapshot_id,
            'Attribute' : 'createVolumePermission'
        }
    elif client_str == "elb" and function_name == "describe_load_balancers":
        return {
            'LoadBalancerNames': [resource_id]
        }
    elif client_str == "elb" and function_name in ["describe_load_balancers", "describe_load_balancer_attributes",\
        "describe_load_balancer_policies"]:
        return {
            'LoadBalancerName': resource_id
        }
    elif client_str == "acm" and function_name == "describe_certificate":
        return {
            'CertificateArn': arn_str
        }
    elif client_str == "cloudformation" and function_name in ["describe_stack_resource",\
        "describe_stack_events", "describe_stacks", "describe_stack_resource_drifts", \
        "get_stack_policy"]:
        return {
            'StackName': resource_id
        }
    elif client_str == "cloudtrail" and function_name == "describe_trails":
        return {
            'trailNameList': [resource_id]
        }
    elif client_str == "cloudtrail" and function_name in ["get_event_selectors",\
        "get_insight_selectors"]:
        return {
            'TrailName': resource_id
        }
    elif client_str == "apigateway" and function_name in ["get_rest_api",\
        "get_documentation_parts", "get_documentation_versions",\
        "get_gateway_responses", "get_models", "get_request_validators",\
        "get_resources", "get_stages"]:
        return {
            'restApiId': resource_id
        }
    elif client_str == "route53" and function_name == "get_hosted_zone":
        return {
            'Id': resource_id
        }
    elif client_str == "route53" and function_name == "list_resource_record_sets":
        return {
            'HostedZoneId': resource_id
        }
    elif client_str == "iam" and function_name in ["get_user", "list_ssh_public_keys", \
        "get_account_summary", "get_account_password_policy", "list_attached_user_policies"]:
        return {
            'UserName': resource_id
        }
    elif client_str == "iam" and function_name == "get_role":
        return {
            'RoleName': resource_id
        }
    elif client_str == "kms" and function_name in ["get_key_rotation_status", "describe_key",]:
        return {
            'KeyId': resource_id
        }
    elif client_str == "dynamodb" and function_name == "describe_table":
        return {
            'TableName': resource_id
        }
    elif client_str == "dynamodb" and function_name == "describe_backup":
        return {
            'BackupArn': arn_str
        }
    elif client_str == "ecs" and function_name == "describe_task_definition":
        return {
            'taskDefinition': resource_id
        }
    elif client_str == "eks" and function_name == "describe_cluster":
        return {
            'name': resource_id
        }
    elif client_str == "elasticache" and function_name == "describe_replication_groups":
        return {
            'ReplicationGroupId': resource_id
        }
    elif client_str == "elasticache" and function_name == "describe_cache_subnet_groups":
        return {
            'CacheSubnetGroupName': resource_id
        }
    elif client_str == "kinesis" and function_name == "describe_stream":
        return {
            'StreamName': resource_id
        }
    elif client_str == "lambda" and function_name == "get_function":
        return {
            'FunctionName': resource_id
        }
    elif client_str == "redshift" and function_name == "describe_clusters":
        return {
            'ClusterIdentifier': resource_id
        }
    elif client_str == "sns" and function_name == "get_topic_attributes":
        return {
            'TopicArn': arn_str
        }
    elif client_str == "sqs" and function_name == "get_queue_attributes":
        return {
            'QueueUrl': 'https:{url}'.format(url=resource_id), 'AttributeNames': ['All']
        }
    elif client_str == "config" and function_name in ["describe_configuration_recorders", "describe_configuration_recorder_status"]:
        return {
            'ConfigurationRecorderNames': [resource_id]
        }
    elif client_str == "es" and function_name == "describe_elasticsearch_domain":
        return {
            'DomainName': resource_id
        }
    elif client_str == "cloudfront" and function_name == "get_distribution":
        return {
            'Id': resource_id
        }
    elif client_str == "ec2" and function_name == "describe_vpn_gateways":
        return {
            'VpnGatewayIds': [resource_id]
        }
    elif client_str == "efs" and function_name == "describe_file_systems":
        return {
            'FileSystemId': resource_id
        }
    elif client_str=='ec2'and function_name == 'describe_route_tables':
        return{
            'RouteTableIds': [resource_id]
        }
    elif client_str=='ec2'and function_name == 'describe_network_acls':
        return{
            'NetworkAclIds': [resource_id]
        }
    elif client_str=='rds'and function_name == 'describe_event_subscriptions':
        return{
            'SubscriptionName': resource_id
        }
    elif client_str=='rds'and function_name == 'describe_db_snapshot_attributes':
        return{
            'DBSnapshotIdentifier': resource_id
        }
    else:
        return {}

def _get_aws_client_data_from_node(node, default_client=None, default_region=None):
    """
    Fetches client name and region from ARN, then from the node, 
    then from the connector.
    """
    aws_region = client_str = None
    arn_str = get_field_value(node, 'arn')
    if arn_str:
        arn_obj = arnparse(arn_str)
        client_str = arn_obj.service
        aws_region = arn_obj.region
    if not client_str:
        client_str = get_field_value(node, 'client')
    if not client_str:
        logger.info("No client type provided in snapshot, using client type from connector")
        client_str = default_client
    if not aws_region: 
        aws_region = get_field_value(node, 'region')
    if not aws_region:
        logger.info("No region provided in snapshot, using region from connector")
        aws_region = default_region
    aws_region = aws_region or default_region
    client_str = client_str or default_client
    return client_str, aws_region


def populate_aws_snapshot(snapshot, container=None):
    """
    This is an entrypoint for populating a snapshot of type aws.
    All snapshot connectors should take snapshot object and based on
    'source' field create a method to connect to the service for the
    connector.
    The 'source' field could be used by more than one snapshot, so the
    'testuser' attribute should match to the user the 'source'
    """
    dbname = config_value('MONGODB', 'dbname')
    snapshot_source = get_field_value(snapshot, 'source')
    snapshot_user = get_field_value(snapshot, 'testUser')
    account_id = get_field_value(snapshot, 'accountId')
    sub_data = get_aws_data(snapshot_source)
    snapshot_nodes = get_field_value(snapshot, 'nodes')
    snapshot_data, valid_snapshotids = validate_snapshot_nodes(snapshot_nodes)
    # valid_snapshotids = True
    # if snapshot_nodes:
    #     for node in snapshot_nodes:
    #         snapshot_data[node['snapshotId']] = False
    #         if not isinstance(node['snapshotId'], str):
    #             valid_snapshotids = False
    # if not valid_snapshotids:
    #     logger.error('All snap')
    if valid_snapshotids and sub_data and snapshot_nodes:
        logger.debug(sub_data)
        access_key, secret_access, region, connector_client_str = \
            get_aws_client_data(sub_data, snapshot_user, account_id)
        if not access_key:
            logger.info("No access_key in the snapshot to access aws resource!...")
            raise Exception("No access_key in the snapshot to access aws resource!...")
            # return snapshot_data

        # Read the client secrets from envirnment variable
        if not secret_access:
            secret_access = os.getenv(snapshot_user, None)
            if secret_access:
                logger.info('Secret Access key from environment variable, Secret: %s', '*' * len(secret_access))
        
        # Read the client secrets from the vault
        if not secret_access:
            secret_access = get_vault_data(access_key)
            if secret_access:
                logger.info('Secret Access key from vault Secret: %s', '*' * len(secret_access))
            elif get_from_currentdata(CUSTOMER):
                logger.error("Secret Access key does not set in a vault")
                raise Exception("Secret Access key does not set in a vault")

        if not secret_access:
            raise Exception("No `secret-access` key in the connector file to access aws resource!...")

        if access_key and secret_access:
            # existing_aws_client = {}
            for node in snapshot['nodes']:
                validate = node['validate'] if 'validate' in node else True
                mastercode = False
                if 'snapshotId' in node and validate:
                    client_str, aws_region = _get_aws_client_data_from_node(node,
                        default_client=connector_client_str, default_region=region)
                    if not _validate_client_name(client_str):
                        logger.error("Invalid Client Name")
                        return snapshot_data
                    try:
                        awsclient = client(client_str.lower(), aws_access_key_id=access_key,
                                           aws_secret_access_key=secret_access, region_name=aws_region)
                    except Exception as ex:
                        logger.info('Unable to create AWS client: %s', ex)
                        awsclient = None
                    logger.info(awsclient)
                    if awsclient:
                        data = get_node(awsclient, node, snapshot_source)
                        if data:
                            error_str = data.pop('error', None)
                            if get_dbtests():
                                if get_collection_size(data['collection']) == 0:
                                    #Creating indexes for collection
                                    create_indexes(data['collection'],
                                        config_value(DATABASE, DBNAME), 
                                        [('snapshotId', pymongo.ASCENDING),
                                        ('timestamp', pymongo.DESCENDING)])

                                    create_indexes(
                                        data['collection'],
                                        config_value(DATABASE, DBNAME), 
                                        [
                                            ('_id', pymongo.DESCENDING),
                                            ('timestamp', pymongo.DESCENDING),
                                            ('snapshotId', pymongo.ASCENDING)
                                        ])
                                check_key = is_check_keys_required(data)
                                insert_one_document(data, data['collection'], dbname, check_key)
                            else:
                                snapshot_dir = make_snapshots_dir(container)
                                if snapshot_dir:
                                    store_snapshot(snapshot_dir, data)
                            if 'masterSnapshotId' in node:
                                snapshot_data[node['snapshotId']] = node['masterSnapshotId']
                            else:
                                snapshot_data[node['snapshotId']] = False if error_str else True
                        else:
                            node['status'] = 'inactive'
                elif 'masterSnapshotId' in node:
                    mastercode = True
                    client_str, aws_region = _get_aws_client_data_from_node(node,
                        default_client=connector_client_str, default_region=region)
                    if not _validate_client_name(client_str):
                        logger.error("Invalid Client Name")
                        return snapshot_data
                    if aws_region:
                        all_regions = [aws_region]
                    else:
                        all_regions = Session().get_available_regions(client_str.lower())
                        if client_str.lower() in ['s3','cloudtrail']:
                            all_regions = ['us-west-1']
                    logger.info("Length of all regions is %s"%(str(len(all_regions))))
                    count = 0
                    snapshot_data[node['masterSnapshotId']] = []
                    for each_region in all_regions:
                        logger.info(each_region)
                        try:
                            awsclient = client(client_str.lower(), aws_access_key_id=access_key,
                                               aws_secret_access_key=secret_access, region_name=each_region)
                        except Exception as ex:
                            logger.info('Unable to create AWS client: %s', ex)
                        logger.info(awsclient)
                        if awsclient:
                            all_data = get_all_nodes(awsclient, node, snapshot, sub_data)
                            if all_data:
                                for data in all_data:
                                    snapshot_data[node['masterSnapshotId']].append(
                                        {
                                            'snapshotId': '%s%s' % (node['masterSnapshotId'], str(count)),
                                            'validate': validate,
                                            'detailMethods': data['detailMethods'],
                                            'structure': 'aws',
                                            'masterSnapshotId': node['masterSnapshotId'],
                                            'collection': data['collection'],
                                            'arn' : data['arn'],
                                            'status' : 'active'
                                        })
                                    count += 1
            if mastercode:
                snapshot_data = eliminate_duplicate_snapshots(snapshot_data)
    return snapshot_data

def is_check_keys_required(data):
    try:
        data = json.dumps(data)
        return True
    except Exception:
        return False

def eliminate_duplicate_snapshots(snapshot_data):
    data = {}
    is_updated = False
    for snapshot_id, value in snapshot_data.items():
        is_updated = False
        for count, snapshot in enumerate(value):
            for sid, sval in data.items():
                for cnt, val in enumerate(sval):
                    if sid == snapshot_id:
                        continue
                    if snapshot['arn'] == val['arn'] and snapshot['detailMethods'] == val['detailMethods']:
                        is_updated = True
                        s_id = snapshot_data[snapshot_id][count]['masterSnapshotId']
                        if isinstance(val['masterSnapshotId'], str):
                            data[sid][cnt]['masterSnapshotId'] = [s_id, val['masterSnapshotId']]
                        elif isinstance(val['masterSnapshotId'], list):
                            data[sid][cnt]['masterSnapshotId'].append(s_id)

        if not is_updated:
            data.update({snapshot_id:value})
    return data


def get_aws_client_data(aws_data, snapshot_user, account_id):
    """
    AWS client information as required by the Boto client, viz access_key
    access_secret, AWS command type like EC2, S3 etc and region
    The access_secret is either read from structure json or env variable or keyvault
    """
    accesskey = None
    secret_access = None
    region = None
    client_str = None
    if aws_data and snapshot_user:
        accounts = get_field_value(aws_data, "accounts")
        if accounts:
            found = False
            for account in accounts:
                if account_id == get_field_value(account, "account-id"):
                    users = get_field_value(account, "users")
                    if users:
                        for user in users:
                            if snapshot_user == get_field_value(user, "name"):
                                found = True
                                accesskey = get_field_value(user, 'access-key')
                                secret_access = get_field_value(user, 'secret-access')
                                region = get_field_value(user, 'region')
                                client_str = get_field_value(user, 'client')
                                if client_str and not _validate_client_name(client_str):
                                    logger.error("Invalid Client Name")
                                break
                if found:
                    break

    return accesskey, secret_access, region, client_str



class MalformedArnError(Exception):
    def __init__(self, arn_str):
        self.arn_str = arn_str

    def __str__(self):
        return 'arn_str: {arn_str}'.format(arn_str=self.arn_str)


class Arn(object):
    def __init__(self, partition, service, region, account_id, resource_type, resource):
        self.partition = partition
        self.service = service
        self.region = region
        self.account_id = account_id
        self.resource_type = resource_type
        self.resource = resource


def arnparse(arn_str):
    if not arn_str.startswith('arn:'):
        raise MalformedArnError(arn_str)

    elements = arn_str.split(':', 5)
    service = elements[2]
    resource = elements[5]

    if service in ['s3', 'sns', 'apigateway', 'execute-api', 'acm', 'rds']:
        resource_type = None
    else:
        resource_type, resource = _parse_resource(resource)

    return Arn(
        partition=elements[1],
        service=service,
        region=elements[3] if elements[3] != '' else None,
        account_id=elements[4] if elements[4] != '' else None,
        resource_type=resource_type,
        resource=resource,
    )


def _parse_resource(resource):
    first_separator_index = -1
    for idx, c in enumerate(resource):
        if c in (':', '/'):
            first_separator_index = idx
            break

    if first_separator_index != -1:
        resource_type = resource[:first_separator_index]
        resource = resource[first_separator_index + 1:]
    else:
        resource_type = None

    return resource_type, resource


import json
from processor.logging.log_handler import getlogger
from processor.helper.json.json_utils import json_from_file
from processor.template_processor.base.base_template_processor import TemplateProcessor
from processor.templates.aws.aws_parser import AWSTemplateParser
from processor.helper.file.file_utils import exists_file
from cfn_flip import flip, to_yaml, to_json

logger = getlogger()

class AWSTemplateProcessor(TemplateProcessor):
    """
    Base Template Processor for process template 
    """

    def __init__(self, node, **kwargs):
        super().__init__(node, tosave=False, **kwargs)
    
    def is_parameter_file(self, file_path):
        """
        check for valid parameter file for parse cloudformation template
        """
        if len(file_path.split(".")) > 0 and file_path.split(".")[-1] in ["json","yaml"]:
            json_data = json_from_file(file_path)
            if json_data and isinstance(json_data, list):
                parameter = json_data[0]
                if isinstance(parameter, dict) and "ParameterKey" in parameter and "ParameterValue" in parameter:
                    return True
        return False

    def is_template_file(self, file_path):
        """
        check for valid template file for parse cloudformation template
        """
        if len(file_path.split(".")) > 0 and file_path.split(".")[-1] in ["json","yaml"]:
            template_json = None
            if file_path.endswith(".yaml") and exists_file(file_path):
                with open(file_path) as yml_file:
                    try:
                        template_json = json.loads(to_json(yml_file.read()))
                    except:
                        pass
            elif file_path.endswith(".json"):
                template_json = json_from_file(file_path)

            if template_json and "AWSTemplateFormatVersion" in template_json:
                return True
        return False

    def process_template(self, paths):
        """
        process the files stored at specified paths and returns the template
        """
        template_json = None
        if paths and isinstance(paths, list):
            parameter_file = None
            template_file = None
            for path in paths:
                file_path = ('%s/%s' % (self.dir_path, path)).replace("//", "/")
                if file_path.endswith("json") and self.is_parameter_file(file_path):
                    parameter_file = file_path
                else:
                    template_file = file_path

            if template_file:
                if parameter_file:
                    aws_template_parser = AWSTemplateParser(template_file, parameter_file=parameter_file)
                else:
                    aws_template_parser = AWSTemplateParser(template_file)
                template_json = aws_template_parser.parse()
        return template_json



"""Helper functions to get data  from KV."""
import json
from urllib.error import HTTPError, URLError
import os
import copy
from urllib import request, parse
from processor.helper.config.config_utils import DBURL, config_value


def json_from_string(json_str):
    """Get json from the string."""
    try:
        jsondata = json.loads(json_str)
        return jsondata
    except:
        pass
    return None


def urlopen_request(urlreq, method):
    """Common utility to trigger the http request."""
    try:
        urlresp = request.urlopen(urlreq)
        respdata = urlresp.read()
        st_code = urlresp.status
        if isinstance(respdata, bytes):
            respdata = respdata.decode()
        data = json_from_string(respdata)
    except HTTPError as ex:
        st_code = ex.code
        data = ex.msg if method == "POST" else None
    except URLError as ex:
        st_code = 500
        data = str(ex)
    return st_code, data


def get_request_headers(headers=None):
    """Add json and no cache headers to the existing headers if passed."""
    hdrs = {
        "Cache-Control": "no-cache",
        "Accept": "application/json"
    }
    req_headers = copy.copy(headers) if headers else {}
    req_headers.update(hdrs)
    return req_headers


def http_get_request(url, headers=None, name='GET'):
    """Get method sends and accepts JSON format."""
    if not url:
        return None, None
    urlreq = request.Request(url, headers=get_request_headers(headers), method=name)
    return urlopen_request(urlreq, name)


def http_post_request(url, mapdata, headers=None, json_type=False, name='POST'):
    """Post method sends and accepts JSON format"""
    if not url:
        return None, None
    if json_type:
        postdata = str.encode(json.dumps(mapdata))
    else:
        postdata = parse.urlencode(mapdata).encode()
    urlreq = request.Request(url, data=postdata, headers=get_request_headers(headers),
                             method='POST')
    return urlopen_request(urlreq, name)


def get_uami_vault_access_token():
    """
    Get the vault access token to get all the other passwords/secrets.
    """
    hdrs = {
       "Metadata": "true",
       "Cache-Control": "no-cache"
    }
    url = 'http://169.254.169.254/metadata/identity/oauth2/token?api-version=2018-02-01&resource=https%3A%2F%2Fvault.azure.net'
    # print('Get Azure UAMI token REST API invoked!')
    status, data = http_get_request(url, headers=hdrs)
    # print(data)
    if status and isinstance(status, int) and status == 200:
        vaulttoken = data['access_token']
    else:
        pass
    return vaulttoken


def get_vault_access_token(tenant_id, vault_client_id, client_secret=None):
    """ Get the vault access token to get all the other passwords/secrets.  """
    vaulttoken = None
    if client_secret:
        vault_client_secret = client_secret if client_secret else None
        data = {
            'grant_type': 'client_credentials',
            'client_id': vault_client_id,
            'client_secret': vault_client_secret,
            'resource': 'https://vault.azure.net'
        }
        hdrs = {
            'Cache-Control': "no-cache",
            "Accept": "application/json"
        }
        if tenant_id:
            url = 'https://login.microsoftonline.com/%s/oauth2/token' % tenant_id
            status, data = http_post_request(url, data, headers=hdrs)
            if status and isinstance(status, int) and status == 200:
                vaulttoken = data['access_token']
    return vaulttoken


def get_keyvault_secret(keyvault, secret_key, vaulttoken):
    hdrs = {
        'Authorization': 'Bearer %s' % vaulttoken
    }
    urlstr = 'https://%s.vault.azure.net/secrets/%s?api-version=7.0'
    url = urlstr % (keyvault, secret_key)
    status, data = http_get_request(url, hdrs)
    return data


def get_azure_vault_data(secret_key=None):
    """Fetches the bearer token for Azure Vault API calls"""
    val = None
    vaulttoken = _get_vault_token()
    if vaulttoken and secret_key:
        keyvault = config_value('VAULT', 'keyvault')
        secret_data = get_keyvault_secret(keyvault, secret_key, vaulttoken)
        if secret_data and 'value' in secret_data:
            val = secret_data['value']
    return val

def _get_vault_token():
    """Fetches the bearer token for Azure Vault API calls"""
    if 'UAMI' in os.environ and os.environ['UAMI'] == 'true':
        vaulttoken = get_uami_vault_access_token()
    else:
        client_id = config_value('VAULT', 'client_id')
        client_secret = get_config_value('VAULT', 'client_secret', 'CLIENTKEY')
        tenant_id = config_value('VAULT', 'tenant_id')
        vaulttoken = get_vault_access_token(tenant_id, client_id, client_secret)
    return vaulttoken

def get_config_value(section, key, env_var, prompt_str=None):
    """ Return the client secret used for the current run"""
    client_secret = config_value(section, key)
    if not client_secret and env_var:
        client_secret = os.getenv(env_var, None)
    if not client_secret and prompt_str:
        key_str = '%s_%s' % (section, key)
        if not client_secret:
            client_secret = input(prompt_str)
            if client_secret:
                pass
                # logger.info('Key:%s, sec:%s', key_str, client_secret)
    return client_secret

def get_dburl():
    unittest = os.getenv('UNITTEST', "false")
    if unittest == 'true':
        return None
    dburl = get_azure_vault_data(DBURL)
    if dburl:
        os.environ[DBURL] = dburl
    return dburl





"""
Define an interface for parsing azure template and its parameter files.
"""
import json
import copy
import os
import glob
import re
import tempfile
import importlib
import sys
from jinja2 import Environment, FileSystemLoader, Undefined, make_logging_undefined
from processor.helper.file.file_utils import save_file, exists_file
from processor.helper.json.json_utils import get_field_value, json_from_file, save_json_to_file
from processor.logging.log_handler import getlogger
from processor.templates.base.template_parser import TemplateParser
from processor.templates.google.util import ResourceContext, SilentUndefined

logger = getlogger()

class GoogleTemplateParser(TemplateParser):
    """
    Google Parser class for process deployment manager template
    """
    def __init__(self, template_file, tosave=False, **kwargs):
        super().__init__(template_file, tosave=False, **kwargs)
        self.imports = []
        self.schema_filter = {
            "integer" : int,
            "boolean" : bool
        }
        self.python_import = False
    
    def update_properties_from_schema_file(self, jinja_file_path):
        """
        search for the schema file and if exist then update the properties with default value
        """
        logger.info("Finding schema file for jinja: %s", jinja_file_path)
        schema_file_path = ("%s.%s") % (jinja_file_path, "schema")
        if exists_file(schema_file_path):
            self.check_imports(schema_file_path)
            f = open(schema_file_path, "r")
            scehma_file = tempfile.mkdtemp()
            yaml_file_path = ("%s/%s") % (scehma_file, "scehma_file.yaml")
            save_file(yaml_file_path, f.read())
            
            schema_json = self.yaml_to_json(yaml_file_path)
            if "properties" in schema_json:
                for key, value in schema_json["properties"].items():
                    if key not in self.gparams and "default" in value:
                        if "type" in value:
                            if value['type'] in self.schema_filter:
                                self.gparams[key] = self.schema_filter[value['type']](value["default"])
                        else:
                            self.gparams[key] = value["default"]
    
    def check_imports(self, json_data):
        """
        check import file and update the import list
        """
        if "imports" in json_data:
            for import_file in json_data["imports"]:
                self.imports.append(import_file)

    def generate_template_json(self):
        """
        generate the template json from template and parameter files
        """
        template_json = self.yaml_to_json(self.get_template())
        gen_template_json = None
        if template_json:
            gen_template_json = copy.deepcopy(template_json)
            self.check_imports(template_json)
            if 'resources' in template_json:
                new_resources = []
                for resource in template_json['resources']:
                    if 'properties' in resource and resource['properties']:
                        self.gparams = resource['properties']
                    else:
                        self.gparams = {}
            
                    processed_resources = self.process_resource(resource)
                    new_resources = new_resources + processed_resources
                
                if not self.python_import:
                    gen_template_json['resources'] = new_resources
                else:
                    gen_template_json = None     
        return gen_template_json

    def process_resource(self, resource):
        """ 
        process the resource json and return the resource with updated values
        """
        new_resources = [resource]
        if "type" in resource and (".jinja" in resource["type"] or ".py" in resource["type"]) and not self.python_import:

            import_file_path = resource["type"]
            for import_file in self.imports:
                if "name" in import_file and import_file["name"] == resource["type"]:
                    import_file_path = import_file["path"]

            full_path_list = self.get_template().split("/")[:-1]
            full_path = ("/".join(full_path_list)).replace("//","/")
            resource_file_path = ("%s/%s" % (full_path, import_file_path)).replace("//","/")

            temp_file_path = "/".join(resource_file_path.split("/")[3:])
            logger.info("Fetch resource from `%s` file", temp_file_path)
            
            if exists_file(resource_file_path):
                self.update_properties_from_schema_file(resource_file_path)

                if ".jinja" in import_file_path:
                    if len(import_file_path.split("/")) > 1:
                        import_file_path_list = import_file_path.split("/")
                        import_file_path = import_file_path_list[-1]
                        full_path = (full_path + "/" + "/".join(import_file_path_list[:-1])).replace("//","/")
                    
                    LoggingUndefined = make_logging_undefined(logger=logger,base=SilentUndefined)
                    env = Environment(loader=FileSystemLoader(full_path), undefined=LoggingUndefined)
                    template = env.get_template(import_file_path)
                    render_data = {
                        "properties" : self.gparams
                    }

                    try:
                        template_render = template.render(**render_data)
                        resource_file = tempfile.mkdtemp()
                        yaml_file_path = ("%s/%s") % (resource_file, "resource_file.yaml")
                        save_file(yaml_file_path, template_render)
                        resource_json = self.yaml_to_json(yaml_file_path)
                    except:
                        resource_json = None

                    if resource_json:
                        self.check_imports(resource_json)
                        if "resources" in resource_json:
                            new_resource_list = []
                            for resource in resource_json["resources"]:
                                if 'properties' in resource and resource["properties"]:
                                    self.gparams.update(resource["properties"])
                                new_resource_list = new_resource_list + self.process_resource(resource)
                            new_resources = new_resource_list

                elif ".py" in import_file_path and exists_file(resource_file_path):
                    # skip the template if any of the resource require to process python file
                    self.python_import = True
                    return new_resources

                    pathname, filename = os.path.split(resource_file_path)
                    sys.path.append(os.path.abspath(pathname))
                    modname = os.path.splitext(filename)[0]

                    try:
                        resource_module = importlib.import_module(modname)
                    except:
                        logger.error("Failed to load module: ", modname)
                        return new_resources

                    resource_context = ResourceContext(self.gparams)
                    genreate_config = getattr(resource_module, "generate_config", None)
                    if not genreate_config:
                        genreate_config = getattr(resource_module, "GenerateConfig", None)
                    if not genreate_config:
                        logger.error("`genreate_config` or `GenerateConfig` method not defined in %s", resource["type"])
                        return new_resources

                    try:
                        resource_json = genreate_config(resource_context)
                        if resource_json:
                            self.check_imports(resource_json)
                            if "resources" in resource_json:
                                new_resource_list = []
                                for resource in resource_json["resources"]:
                                    if 'properties' in resource and resource["properties"]:
                                        self.gparams.update(resource["properties"])
                                    new_resource_list = new_resource_list + self.process_resource(resource)
                                new_resources = new_resource_list
                    except Exception as e:
                        logger.error("Failed to fetch resource. Missing parmater : %s", str(e))
        return new_resources


"""
   Filesystem snapshot specific functionality using the base methods.
"""
import json
from processor.helper.file.file_utils import exists_file
from processor.logging.log_handler import getlogger
from processor.helper.json.json_utils import get_field_value, json_from_file,\
    TEST, MASTERTEST, SNAPSHOT, MASTERSNAPSHOT, make_snapshots_dir,\
    store_snapshot, get_field_value_with_default, get_json_files, get_container_snapshot_json_files
from processor.helper.config.config_utils import config_value, framework_dir, get_test_json_dir
from processor.connector.snapshot_base import Snapshot
from processor.connector.snapshot_exception import SnapshotsException

logger = getlogger()


class FSSnapshot(Snapshot):
    """
    Filesystem snapshot utilities.
    """
    def __init__(self, container, snapshot_refactored_fns, singleTest=None):
        """ Default isDb is false, singletest shall be set to the test that needs to be run."""
        super().__init__(container, snapshot_refactored_fns)
        self.singleTest = singleTest
        reporting_path = config_value('REPORTING', 'reportOutputFolder')
        self.container_dir = '%s/%s/%s' % (framework_dir(), reporting_path, container)

    def get_structure_data(self, snapshot_object):
        """ Get the structure from the filesystem."""
        structure_data = {}
        json_test_dir = get_test_json_dir()
        snapshot_source = get_field_value(snapshot_object, "source")
        file_name = '%s.json' % snapshot_source if snapshot_source and not \
            snapshot_source.endswith('.json') else snapshot_source
        custom_source = '%s/../%s' % (json_test_dir, file_name)
        logger.info('%s structure file is %s', Snapshot.LOGPREFIX, custom_source)
        if exists_file(custom_source):
            structure_data = json_from_file(custom_source)
        return structure_data

    def get_used_snapshots_in_tests(self):
        """ Iterate through all snapshot and mastersnapshot and list the used snapshots in tests or mastertest."""
        snapshots = []
        logger.info("%s Fetching files for %s from container dir: %s", Snapshot.LOGPREFIX, self.container,
                    self.container_dir)
        for testType, snapshotType, replace in (
                (TEST, SNAPSHOT, False),
                (MASTERTEST, MASTERSNAPSHOT, True)):
            test_files = get_json_files(self.container_dir, testType)
            logger.info('%s fetched %s number of files from %s container: %s', Snapshot.LOGPREFIX, snapshotType,
                        self.container, len(test_files))
            snapshots.extend(self.process_files(test_files, snapshotType, replace))
        return list(set(snapshots))


    def process_files(self, test_files, doctype, replace=False):
        """ Process Test or masterTest json files."""
        snapshots = []
        for test_file in test_files:
            test_json_data = json_from_file(test_file)
            if test_json_data:
                snapshot = test_json_data[doctype] if doctype in test_json_data else ''
                if snapshot:
                    file_name = snapshot if snapshot.endswith('.json') else '%s.json' % snapshot
                    if replace:
                        file_name = file_name.replace('.json', '_gen.json')
                    if self.singleTest:
                        testsets = get_field_value_with_default(test_json_data, 'testSet', [])
                        for testset in testsets:
                            for testcase in testset['cases']:
                                if ('testId' in testcase and testcase['testId'] == self.singleTest) or \
                                        ('masterTestId' in testcase and testcase['masterTestId'] == self.singleTest):
                                    if file_name not in snapshots:
                                        snapshots.append(file_name)
                    else:
                        snapshots.append(file_name)
        return snapshots

    def store_data_node(self, data):
        """Store the data in the filesystem"""
        # Make a snapshots directory if DB is NONW
        snapshot_dir = make_snapshots_dir(self.container)
        if snapshot_dir:
            store_snapshot(snapshot_dir, data)

    def get_snapshots(self):
        """
        Get the snapshot files from the container with storage system as filesystem.
        The path for looking into the container is configured in the config.ini, for the
        default location configuration is $SOLUTIONDIR/relam/validation/<container>
        """
        snapshots_status = {}
        snapshot_dir, snapshot_files = get_container_snapshot_json_files(self.container)
        if not snapshot_files:
            logger.error("%s No Snapshot for this container: %s, in %s, add and run again!...", Snapshot.LOGPREFIX, self.container, snapshot_dir)
            raise SnapshotsException("No snapshots for this container: %s, add and run again!..." % self.container)
        used_snapshots = self.get_used_snapshots_in_tests()
        populated = []
        for snapshot_file in snapshot_files:
            parts = snapshot_file.rsplit('/', 1)
            if parts[-1] in used_snapshots and parts[-1] not in populated:
                # Take the snapshot and populate whether it was successful or not.
                # Then pass it back to the validation tests, so that tests for those
                # snapshots that have been susccessfully fetched shall be executed.
                file_name = '%s.json' % snapshot_file if snapshot_file and not snapshot_file.endswith('.json') else snapshot_file
                snapshot_json_data = json_from_file(file_name)
                if not snapshot_json_data:
                    logger.info("%s snapshot file %s looks to be empty, next!...",  Snapshot.LOGPREFIX, snapshot_file)
                    continue

                pull_response, git_connector_json = self.check_and_fetch_remote_snapshots(snapshot_json_data)
                if git_connector_json and not pull_response:
                    logger.info('%s Fetching remote snapshots failed.', Snapshot.LOGPREFIX)
                    break

                logger.debug(json.dumps(snapshot_json_data, indent=2))
                snapshot_data = self.populate_snapshots(snapshot_json_data)
                populated.append(parts[-1])
                name = parts[-1].replace('.json', '') if parts[-1].endswith('.json') else parts[-1]
                snapshots_status[name] = snapshot_data
        return snapshots_status



"""All comparison functions."""

import math

EQ = '='
NEQ = '!='
GT = '>'
GTE = '>='
LT = '<'
LTE = '<='


int_funcs = {
    EQ: lambda lhs, rhs: lhs == rhs,
    NEQ: lambda lhs, rhs: lhs != rhs,
    GT: lambda lhs, rhs: lhs > rhs,
    GTE: lambda lhs, rhs: lhs >= rhs,
    LT: lambda lhs, rhs: lhs < rhs,
    LTE: lambda lhs, rhs: lhs <= rhs
}

float_funcs = {
    EQ: lambda lhs, rhs: math.isclose(lhs, rhs),
    NEQ: lambda lhs, rhs: not math.isclose(lhs, rhs),
    GT: lambda lhs, rhs: lhs > rhs,
    GTE: lambda lhs, rhs: lhs > rhs or math.isclose(lhs, rhs),
    LT: lambda lhs, rhs: lhs < rhs,
    LTE: lambda lhs, rhs: lhs < rhs or math.isclose(lhs, rhs)
}


def compare_none(loperand, roperand, op):
    if loperand is None and roperand is None:
        return True if op == EQ else False
    return False


def compare_int(loperand, roperand, op):
    if type(loperand) is int and type(roperand) is int:
        if op in int_funcs:
            return int_funcs[op](loperand, roperand)
    return False


def compare_float(loperand, roperand, op):
    if type(loperand) is float and type(roperand) is float:
        if op in float_funcs:
            return float_funcs[op](loperand, roperand)
    return False


def compare_boolean(loperand, roperand, op):
    if type(loperand) is bool and type(roperand) is bool:
        if op == EQ:
            return True if loperand == roperand else False
        elif op == NEQ:
            return True if loperand != roperand else False
    return False


def compare_str(loperand, roperand, op):
    if type(loperand) is str and type(roperand) is str:
        if op in int_funcs:
            return int_funcs[op](loperand, roperand)
    return False


def compare_list(loperand, roperand, op):
    if type(loperand) is list and type(roperand) is list:
        if op in int_funcs:
            return int_funcs[op](loperand, roperand)
    return False

def compare_in(loperand, roperand, op):
    if loperand and roperand:
        return roperand in loperand
    return False

def compare_dict(loperand, roperand, op):
    if type(loperand) is dict and type(roperand) is dict:
        if op in int_funcs:
            return int_funcs[op](loperand, roperand)
    return False



from processor.template_processor.aws_template_processor import AWSTemplateProcessor
from processor.template_processor.azure_template_processor import AzureTemplateProcessor
from processor.template_processor.google_template_processor import GoogleTemplateProcessor

TEMPLATE_NODE_TYPES = {
    "cloudformation": AWSTemplateProcessor,
    "arm" : AzureTemplateProcessor,
    "deploymentmanager" : GoogleTemplateProcessor,
}


"""
   Filesystem snapshot specific functionality using the base methods.
"""
import pymongo
from processor.logging.log_handler import getlogger
from processor.helper.json.json_utils import get_field_value, collectiontypes, \
    STRUCTURE, TEST, MASTERTEST, SNAPSHOT, MASTERSNAPSHOT
from processor.helper.config.config_utils import config_value
from processor.database.database import insert_one_document, get_collection_size, create_indexes, \
     DATABASE, DBNAME, sort_field, get_documents, update_one_document
from processor.reporting.json_output import dump_output_results
from processor.connector.snapshot_base import Snapshot
from processor.connector.snapshot_exception import SnapshotsException


logger = getlogger()


class DBSnapshot(Snapshot):
    """
    Database snapshot utilities.
    """
    def __init__(self, container, snapshot_refactored_fns):
        """"DB is true, will be usefule to make checks."""
        super().__init__(container)
        self.dbname = config_value(DATABASE, DBNAME)
        self.qry = {'container': container}
        self.sort = [sort_field('timestamp', False)]
        self.isDb = True

    def collection(self, name=TEST):
        """ Get the collection name for the json object"""
        return config_value(DATABASE, collectiontypes[name])

    def get_structure_data(self, snapshot_object):
        """ Return the structure from the database"""
        structure_data = {}
        snapshot_source = get_field_value(snapshot_object, "source")
        snapshot_source = snapshot_source.replace('.json', '') if snapshot_source else ''
        qry = {'name': snapshot_source}
        structure_docs = get_documents(self.collection(STRUCTURE), dbname=self.dbname, sort=self.sort, query=qry, limit=1)
        logger.info('%s fetched %s number of documents: %s', Snapshot.LOGPREFIX, STRUCTURE, len(structure_docs))
        if structure_docs and len(structure_docs):
            structure_data = structure_docs[0]['json']
        return structure_data

    def get_used_snapshots_in_tests(self):
        """ Get the snapshots used in test and mastertest of the container."""
        snapshots = []
        logger.info("%s Fetching documents for %s", Snapshot.LOGPREFIX, self.container)
        for collection, snapshotType, suffix in (
                (TEST, SNAPSHOT, ''),
                (MASTERTEST, MASTERSNAPSHOT, '_gen')):
            docs = get_documents(self.collection(collection), dbname=self.dbname, sort=self.sort, query=self.qry)
            logger.info('%s fetched %s number of documents: %s', Snapshot.LOGPREFIX, collection, len(docs))
            snapshots.extend(self.process_docs(docs, snapshotType, suffix))
        return list(set(snapshots))

    def get_snapshots(self):
        """Populate the used snapshots in test and mastertest for this container."""
        snapshots_status = {}
        docs = get_documents(self.collection(SNAPSHOT), dbname=self.dbname, sort=self.sort, query=self.qry, _id=True)
        if docs and len(docs):
            logger.info('%s fetched %s number of documents: %s', Snapshot.LOGPREFIX, SNAPSHOT, len(docs))
            used_snapshots = self.get_used_snapshots_in_tests()
            if not used_snapshots:
                raise SnapshotsException("No snapshots for this container: %s, add and run again!..." % self.container)
            populated = []
            for doc in docs:
                if doc['json']:
                    snapshot = doc['name']
                    try:
                        pull_response, git_connector_json = self.check_and_fetch_remote_snapshots(doc['json'])
                        if git_connector_json and not pull_response:
                            logger.info('%s Fetching remote snapshots failed.', Snapshot.LOGPREFIX)
                            break

                        if snapshot in used_snapshots and snapshot not in populated:
                            # Take the snapshot and populate whether it was successful or not.
                            # Then pass it back to the validation tests, so that tests for those
                            # snapshots that have been susccessfully fetched shall be executed.
                            snapshot_file_data = self.populate_snapshots(doc['json'])

                            if not git_connector_json:
                                update_one_document(doc, self.collection(SNAPSHOT), self.dbname)

                            populated.append(snapshot)
                            snapshots_status[snapshot] = snapshot_file_data
                    except Exception as e:
                        dump_output_results([], self.container, "-", snapshot, False)
                        raise e
        if not snapshots_status:
            raise SnapshotsException("No snapshots for this container: %s, add and run again!..." % self.container)
        return snapshots_status

    def process_docs(self, docs, doctype, suffix=''):
        """ Process Test or masterTest documents"""
        snapshots = []
        if docs and len(docs):
            for doc in docs:
                if doc['json']:
                    snapshot = doc['json'][doctype] if doctype in doc['json'] else ''
                    if snapshot:
                        snapshots.append((snapshot.split('.')[0] if snapshot.endswith('.json') else snapshot) + suffix)
        return snapshots

    def store_data_node(self, data):
        """ Store to database"""
        if get_collection_size(data['collection']) == 0:
            # Creating indexes for collection
            create_indexes(
                data['collection'],
                config_value(DATABASE, DBNAME),
                [
                    ('snapshotId', pymongo.ASCENDING),
                    ('timestamp', pymongo.DESCENDING)
                ]
            )

            create_indexes(
                data['collection'],
                config_value(DATABASE, DBNAME),
                [
                    ('_id', pymongo.DESCENDING),
                    ('timestamp', pymongo.DESCENDING),
                    ('snapshotId', pymongo.ASCENDING)
                ]
            )
        insert_one_document(data, data['collection'], self.dbname, check_keys=False)



"""
   Base snapshot for filesystem, db etc.
"""
import hcl

from processor.logging.log_handler import getlogger
from processor.helper.json.json_utils import get_field_value, json_from_file
from processor.connector.populate_json import pull_json_data
from processor.helper.yaml.yaml_utils import yaml_from_file


logger = getlogger()


class Snapshot:
    """ Base class for snapshot processing"""

    LOGPREFIX = 'Snapshots:'
    # snapshot_fns = {
    #     'azure1': populate_snapshot_azure,
    #     'azure': populate_azure_snapshot,
    #     'filesystem': populate_custom_snapshot
    # }

    def __init__(self, container, snapshot_refactored_fns):
        """ Base class, where all attributes are false."""
        self.container = container
        self.snapshot_fns = snapshot_refactored_fns
        self.appObject = {}
        self.singleTest = None
        self.isDb = False

    def store_value(self, key, value):
        """ Store key value used down the p[rocessing stream."""
        if key and value:
            self.appObject[key] = value

    def get_value(self, key):
        """ Use the app storage and will be garbage collected after the completion of the snapshot fetch process."""
        if key and key in self.appObject:
            return self.appObject.get(key)
        return None

    def get_snapshots(self):
        """ Iterator based implementation"""
        return []

    def get_snapshot_nodes(self, snapshot):
        """ Iterate over the nodes of the snapshot object"""
        snapshot_nodes = get_field_value(snapshot, 'nodes')
        return snapshot_nodes if snapshot_nodes else []

    def validate_snapshot_ids_in_nodes(self, snapshot):
        """ The snapshotsIds should be strings and also quoted."""
        snapshot_data = {}
        valid_snapshotids = True
        for node in self.get_snapshot_nodes(snapshot):
            if 'snapshotId' in node and node['snapshotId']:
                snapshot_data[node['snapshotId']] = False
                if not isinstance(node['snapshotId'], str):
                    valid_snapshotids = False
            elif 'masterSnapshotId' in node and node['masterSnapshotId']:
                snapshot_data[node['masterSnapshotId']] = False
                if not isinstance(node['masterSnapshotId'], str):
                    valid_snapshotids = False
            else:
                logger.error(
                    'All snapshot nodes should contain snapshotId or masterSnapshotId attribute with a string value')
                valid_snapshotids = False
                break
        if not valid_snapshotids:
            logger.error('All snapshot Ids should be strings, even numerals should be quoted')
        return snapshot_data, valid_snapshotids

    def check_and_fetch_remote_snapshots(self, json_data):
        """Could be snapshot that snapshots are fetched from remote repository."""
        git_connector_json = False
        pull_response = False
        if "connector" in json_data and "remoteFile" in json_data and \
                json_data["connector"] and json_data["remoteFile"]:
            git_connector_json = True
            _, pull_response = pull_json_data(json_data)
        return pull_response, git_connector_json

    def get_structure_data(self, snapshot_object):
        """ Return a empty dict in base class."""
        structure_data = {}
        return structure_data

    def store_data_node(self, data):
        """ Store the data record as per the data system"""
        return False

    def populate_snapshots(self, snapshot_json_data):
        """
        Every snapshot should have collection of nodes which are to be populated.
        Each node in the nodes list of the snapshot shall have a unique id in this
        container so as not to clash with other node of a snapshots.
        """
        snapshot_data = {}
        snapshots = get_field_value(snapshot_json_data, 'snapshots')
        if not snapshots:
            logger.error("Json Snapshot does not contain snapshots, next!...")
            return snapshot_data
        for snapshot in snapshots:
            connector_data = self.get_structure_data(snapshot)
            snapshot_type = get_field_value(connector_data, "type")
            if snapshot_type and snapshot_type in self.snapshot_fns:
                if 'nodes' not in snapshot or not snapshot['nodes']:
                    logger.error("No nodes in snapshot to be backed up!...")
                    return snapshot_data
                if snapshot_type == 'azure' or snapshot_type == 'filesystem':
                    current_data = self.snapshot_fns[snapshot_type](snapshot, self.container)
                else:
                    current_data = self.snapshot_fns[snapshot_type](snapshot, self)
                logger.info('Snapshot: %s', current_data)
                snapshot_data.update(current_data)
        return snapshot_data

    @classmethod
    def convert_to_json(cls, file_path, file_type):
        json_data = {}
        if file_type == 'json':
            json_data = json_from_file(file_path, escape_chars=['$'])
        elif file_type == 'terraform':
            with open(file_path, 'r') as fp:
                json_data = hcl.load(fp)
        elif file_type == 'yaml' or file_type == 'yml':
            json_data = yaml_from_file(file_path)
        else:
            logger.error("Fileconversion error type:%s and file: %s", file_type, file_path)
        return json_data



"""
Define an interface for parsing azure template and its parameter files.
"""
import json
import copy
import os
import glob
import re
from processor.helper.json.json_utils import get_field_value, json_from_file, save_json_to_file
from processor.logging.log_handler import getlogger
from processor.templates.base.template_parser import TemplateParser

logger = getlogger()

class AzureTemplateParser(TemplateParser):
    """
    Azure Parser class for process Azure arm template
    """
    def __init__(self, template_file, tosave=False, **kwargs):
        super().__init__(template_file, tosave=False, **kwargs)
        self.gvariables = {}
        self.function_handlers = {
            "parameters": self.handle_params,
            "variables": self.handle_variables,
            "concat": self.handle_concat,
            "equals": self.handle_equals,
            "length": lambda x: (True, len(x))
        }
    
    def generate_template_json(self):
        """
        generate the template json from template and parameter files
        """
        stars = '*' * 25
        template_json = json_from_file(self.get_template())
        self.replace_spacial_characters(template_json)
        gen_template_json = None
        if template_json:
            gen_template_json = copy.deepcopy(template_json)
            if 'parameters' not in template_json:
                template_json['parameters'] = {}
            self.gparams = template_json['parameters']
            for param in self.get_parameter():
                param_json = json_from_file(param)
                if 'parameters' in  param_json and param_json['parameters']:
                    for key, value in param_json['parameters'].items():
                        # if key in template_json['parameters']:
                        if "value" in value:
                            if key not in template_json['parameters']:
                                template_json['parameters'][key] = {'value': value['value']}
                            else:
                                template_json['parameters'][key]['value'] = value['value']
                        else:
                            logger.error("From parameter %s was not replaced.", key)
                gen_template_json['parameters'] = self.gparams
            # print('%s Updated Parameters %s' % (stars, stars))
            # print(json.dumps(template_json['parameters'], indent=2))
            if 'variables' in template_json:
                self.gvariables = template_json['variables']
                new_resource = self.process_resource(template_json['variables'])
                # print('%s Original Variables %s' % (stars, stars))
                # print(json.dumps(template_json['variables'], indent=2))
                # print('%s Updated Variables %s' % (stars, stars))
                # print(json.dumps(new_resource, indent=2))
                # Second pass, becoz some variables have been defined in terms of parameters, functions like concat, substr
                self.gvariables = self.process_resource(new_resource)
                gen_template_json['variables'] = self.gvariables
            if 'resources' in template_json:
                new_resources = []
                for resource in template_json['resources']:
                    is_copy, copy_resources = self.handle_copy(resource)
                    if is_copy:
                        for resourc in copy_resources:
                            new_resource = self.process_resource(resourc)
                            new_resources.append(new_resource)
                    else:
                        new_resource = self.process_resource(resource)
                        new_resources.append(new_resource)
                    # print('%s Original Resource %s' % (stars, stars))
                    # print(json.dumps(resource, indent=2))
                    # print('%s Updated Resource %s' % (stars, stars))
                    # print(json.dumps(new_resource, indent=2))
                gen_template_json['resources'] = new_resources
        return gen_template_json

    def process_resource(self, resource):
        """ 
        process the resource json and return the resource with updated values
        """
        new_resource = resource
        if isinstance(resource ,dict):
            new_resource = {}
            for key, value in resource.items():
                if key == "accessPolicies":
                    print("Here")
                # if key == 'dnsNameForPublicIP':
                #     print("Here")
                if isinstance(value, str):
                    new_resource[key] = value
                    eval_expr = self.eval_expression(value)
                    if eval_expr:
                        func_name, func_params = self.func_details(eval_expr)
                        if func_name in self.function_handlers:
                            success, updated_value = self.function_handlers[func_name](func_params)
                            # print('*' * 50)
                            # print(value)
                            # print(updated_value)
                            # print('#' * 50)
                            new_resource[key] = updated_value if success else value
                else:
                    new_resource[key] = self.process_resource(value)
        elif isinstance(resource ,list):
            success = True
            new_resource = []
            for value in resource:
                if isinstance(value, str):
                    eval_expr = self.eval_expression(value)
                    if eval_expr:
                        func_name, func_params = self.func_details(eval_expr)
                        if func_name in self.function_handlers:
                            lsuccess, updated_value = self.function_handlers[func_name](func_params)
                            # print('*' * 50)
                            # print(value)
                            # print(updated_value)
                            # print('#' * 50)
                            if not lsuccess:
                                success = lsuccess
                            new_resource.append(updated_value)
                        else:
                            new_resource.append(value)
                    else:
                        new_resource.append(value)
                else:
                    new_resource.append(self.process_resource(value))
            if not success:
                new_resource = resource
        return new_resource
    
    def my_split(self, value):
        vals = []
        t = []
        count = 0
        for i in value:
            t.append(i)
            if i == '(':
                count += 1
            elif i == ')':
                count -= 1
            elif i == ',':
                if count == 0:
                    t.pop()
                    vals.append(''.join(t))
                    t = []
        if t:
            vals.append(''.join(t))
        return vals

    def do_parentheses_match(self, input_string):
        s = []
        balanced = True
        index = 0
        while index < len(input_string) and balanced:
            token = input_string[index]
            if token == "(":
                s.append(token)
            elif token == ")":
                if len(s) == 0:
                    balanced = False
                else:
                    s.pop()
            index += 1
        return balanced and len(s) == 0

    def handle_variables(self, variables_expr):
        # print(json.dumps(gvariables, indent=2))
        # print('@' * 50)
        # print(variables_expr)
        val = variables_expr.strip().replace("'", "")
        # if val == 'diagStorageAccName':
        #     print(val)
        if val in self.gvariables:
            # print(json.dumps(gvariables[val], indent=2))
            # return True, gvariables[val]
            success = False
            value = self.gvariables[val]
            if isinstance(value, str):
                eval_expr = self.eval_expression(value)
                if eval_expr:
                    func_name, func_params = self.func_details(eval_expr)
                    if func_name in self.function_handlers:
                        success, updated_value = self.function_handlers[func_name](func_params)
                        return success, updated_value
            return True, value
        else:
            print("%s variable does not exist" % val)
        return True, val


    def handle_params(self, params_expr):
        # print(json.dumps(gparams, indent=2))
        # print('@' * 50)
        # print(params_expr)
        ex_params = None
        exmatch = re.match(r'^(\(.*\))(.*)', params_expr, re.I)
        if exmatch:
            field, ex_params = exmatch.groups()
            val = field[1:-1].strip().replace("'", "")
        else:
            val = params_expr.strip().replace("'", "")
        # print(val)
        if val in self.gparams:
            # print(json.dumps(gparams[val], indent=2))
            if 'value' in self.gparams[val]:
                if ex_params:
                    return True, get_field_value(self.gparams[val]['value'], ex_params)
                return True, self.gparams[val]['value']
            elif 'defaultValue' in self.gparams[val]:
                if ex_params:
                    return True, get_field_value(self.gparams[val]['defaultValue'], ex_params)
                return True, self.gparams[val]['defaultValue']
        else:
            print("%s does not exist" % val)
        return True, val

    def handle_concat(self, concat_expr):
        # values = concat_expr.split(',')
        values = self.my_split(concat_expr)
        if values:
            success = True
            updated_values = []
            for value in values:
                func_name, func_params = self.func_details(value.strip())
                if func_name and func_params: 
                    # print(value)
                    # print(func_name, func_params)
                    lsuccess, updated_value = self.eval_func(func_name, func_params)
                    # print(updated_value)
                    if not lsuccess:
                        success = False
                    updated_values.append(updated_value if updated_value else value)
                else:
                    updated_values.append(value.strip().replace("'", ""))
            # print(updated_values)
            return(success, ''.join(str(value) for value in updated_values) if success else concat_expr)
        return False, concat_expr

    def handle_equals(self, equals_expr):
        values = equals_expr.split(',')
        if values and len(values) == 2:
            updated_values = []
            for value in values:
                func_name, func_params = self.func_details(value)
                if func_name and func_params: 
                    updated_value = self.eval_func(func_name, func_params)
                    updated_values.append(updated_value if updated_value else value)
                else:
                    updated_values.append(value.strip().replace("'", ""))
            # print(updated_values)
            return(True, updated_values[0] == updated_values[1])
        return True, False


    def eval_func(self, func_name, func_details):
        if func_name in self.function_handlers:
            return self.function_handlers[func_name](func_details)
        return False, None
    
    def version_str(self, version):
        """Convert numeric version to string eg: 0.1 => 0_1"""
        value = version.replace('.', '_') if version else version
        return value

    def is_parameter_file(self, filename):
        json_data = json_from_file(filename)
        if json_data and '$schema' in json_data and json_data['$schema']:
            match =  re.match(r'.*deploymentParameters.json#', json_data['$schema'], re.I)
            return True if match else False
        return None
            
    def is_template_file(self, filename):
        if filename and filename.endswith('_gen.json'):
            return False
        json_data = json_from_file(filename)
        if json_data and '$schema' in json_data and json_data['$schema']:
            match =  re.match(r'.*deploymentTemplate.json#$', json_data['$schema'], re.I)
            return True if match else False
        return None
            
    def eval_expression(self, value):
        match = re.match(r'\[(.*)\]', value, re.I)
        if match:
            return match.groups()[0]
        return None

    def func_details(self, value):
        # match = re.match(r'([a-z0-9A-Z]{0,})\((.*)\)', value, re.I)
        match = re.match(r'([a-z0-9A-Z]{0,})(\(.*)', value, re.I)
        if match:
            # Support only plain braces, no indexing
            params = match.groups()[1]
            parammatch = re.match(r'^\(.*\)$', params, re.I)
            exmatch = re.match(r'^(\(.*\))(.*)', params, re.I)
            if parammatch:
                if ('[' not in params or ']' not in params):
                    return match.groups()[0], params[1:-1]
            elif exmatch:
                #if ('[' not in params or ']' not in params):
                return match.groups()[0], params
        return None, None

    def replace_spacial_characters(self, gen_template_json):
        if gen_template_json.get('$schema', None):
            gen_template_json["\uFF04schema"] = gen_template_json["$schema"]
            del gen_template_json["$schema"]

        for key, value in gen_template_json.items():
            if isinstance(value, dict):
                self.replace_spacial_characters(value)
            elif isinstance(value, list):
                for val in value:
                    if isinstance(val, dict):
                        self.replace_spacial_characters(val)

    def handle_copy(self, resource):
        is_copy =  None
        value = None
        if 'copy' in resource:
            eval_expr = self.eval_expression(resource['copy']['count'])
            if eval_expr:
                func_name, func_params = self.func_details(eval_expr)
                if func_name in self.function_handlers:
                    ifunc_name, ifunc_params = self.func_details(func_params)
                    if ifunc_name in self.function_handlers:
                        _, ivalue = self.function_handlers[ifunc_name](ifunc_params)
                        if ivalue:
                            _, value = self.function_handlers[func_name](ivalue)
                            is_copy = resource['copy']['name']
            if is_copy and isinstance(value, int):
                resources = []
                for i in range(value):
                    resource_str = json.dumps(resource)
                    match_str = "copyIndex('%s')" % is_copy
                    resource_str = resource_str.replace(match_str, '%d' % i)
                    new_resource = json.loads(resource_str)
                    del new_resource['copy']
                    resources.append(new_resource)
                return True, resources
        return False, resource


import uuid

from processor.connector.vault import get_vault_data, set_vault_data
from processor_enterprise.controller.vaultcontroller import set_key_visbility, EDITABLE

def generate_password():
	return str(uuid.uuid4())

def generate_azure_vault_key():

	key = input("Enter the key to add or update its password: ")

	is_key_exists = get_vault_data(secret_key=key)

	password = generate_password()

	is_created = set_vault_data(key_name=key, value=password)

	if is_key_exists and is_created:
		print("Regenerating password for key: ", key)
	elif is_created:
		set_key_visbility(key, EDITABLE)
		print("Creating and generating password for key: ", key)
	else:
		print("Getting issue while generating key:", key)




"""
Snapshot utils contains common functionality for all snapshots.
"""
import time
import hashlib
from processor.database.database import COLLECTION, get_documents
from processor.logging.log_handler import getlogger


logger = getlogger()


def validate_snapshot_nodes(snapshot_nodes):
    snapshot_data = {}
    valid_snapshotids = True
    if snapshot_nodes:
        for node in snapshot_nodes:
            if 'snapshotId' in node and  node['snapshotId']:
                snapshot_data[node['snapshotId']] = False
                if not isinstance(node['snapshotId'], str):
                    valid_snapshotids = False
            elif 'masterSnapshotId' in node and node['masterSnapshotId']:
                snapshot_data[node['masterSnapshotId']] = False
                if not isinstance(node['masterSnapshotId'], str):
                    valid_snapshotids = False
            else:
                logger.error('All snapshot nodes should contain snapshotId or masterSnapshotId attribute with a string value')
                valid_snapshotids = False
                break
            # snapshot_data[node['snapshotId']] = False
            # if not isinstance(node['snapshotId'], str):
            #     valid_snapshotids = False
    if not valid_snapshotids:
        logger.error('All snapshot Ids should be strings, even numerals should be quoted')
    return snapshot_data, valid_snapshotids


def get_data_record(ref_name, node, user, snapshot_source, connector_type):
    """ The data node record, common function across connectors."""
    collection = node['collection'] if 'collection' in node else COLLECTION
    parts = snapshot_source.split('.')
    return {
        "structure": connector_type,
        "reference": ref_name,
        "source": parts[0],
        "path": '',
        "timestamp": int(time.time() * 1000),
        "queryuser": user,
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "node": node,
        "snapshotId": node['snapshotId'] if 'snapshotId' in node else '',
        "mastersnapshot": False,
        "masterSnapshotId": node['masterSnapshotId'] if 'masterSnapshotId' in node else '',
        "collection": collection.replace('.', '').lower(),
        "json": {}  # Refactor when node is absent it should None, when empty object put it as {}
    }


"""Utility functions for file and  directory"""

import os
from processor.logging.log_handler import getlogger

logger = getlogger()

def exists_dir(dirname):
    """Check if this path exists and is a directory"""
    if dirname and os.path.exists(dirname) and os.path.isdir(dirname):
        return True
    return False


def exists_file(fname):
    """Check if path exists and is a file"""
    if fname and os.path.exists(fname) and os.path.isfile(fname):
        return True
    return False


def remove_file(fname):
    """Remove the file."""
    try:
        os.remove(fname)
        return True
    except:
        return False


def mkdir_path(dirpath):
    """Make directories recursively."""
    try:
        os.makedirs(dirpath)
        return exists_dir(dirpath)
    except:
        return False

def save_file(file_path, content):
    "write content in file and save file to specified path"
    try:
        f = open(file_path, "w")
        f.write(content)
        f.close()
        return True
    except Exception as e:
        logger.error(e)    
        return False
    


"""
   Common file for populating json files to json.
"""
import argparse
import sys
import atexit
import time
import hashlib
import glob
import json
from pymongo import TEXT
from processor.logging.log_handler import getlogger
from processor.helper.config.rundata_utils import init_currentdata, delete_currentdata
from processor.database.database import init_db, create_indexes, insert_one_document
from processor.helper.config.config_utils import DATABASE, DBNAME, config_value
from processor.helper.file.file_utils import exists_dir, exists_file
from processor.helper.json.json_utils import json_from_file,collectiontypes


logger = getlogger()



def json_record(container, filetype, filename, json_data=None):
    parts = filename.rsplit('/')
    fparts = parts[-1].split('.')
    db_record = {
        "timestamp": int(time.time() * 1000),
        "container": container,
        "checksum": hashlib.md5("{}".encode('utf-8')).hexdigest(),
        "type": filetype,
        "name": fparts[0],
        "collection": config_value(DATABASE, collectiontypes[filetype]),
        "json": json_data if json_data else {}
    }
    if '$schema' in db_record['json']:
        del db_record['json']['$schema']
    return db_record


def populate_json_files(args):
    dbname = config_value(DATABASE, DBNAME)
    if args.dir:
        logger.info("Checking this directory: %s for json files", args.dir)
        json_dir = args.dir
        if exists_dir(args.dir):
            for filename in glob.glob('%s/*.json' % json_dir.replace('//', '/')):
                json_data = json_from_file(filename)
                if json_data and 'fileType' in json_data:
                    filetype = json_data['fileType']
                else:
                    filetype = 'structure'
                logger.info('Storing file:%s from directory: %s', json_dir, filename)
                db_record = json_record(args.container, filetype, filename, json_data)
                if validate_json_data(db_record['json'], db_record['type']):
                    insert_one_document(db_record, db_record['collection'], dbname, False)
                    logger.debug('DB Record: %s', json.dumps(db_record, indent=2))
                else:
                    logger.info('Invalid json for type:%s', db_record['type'])
                logger.info('*' * 80)
    if args.file:
        logger.info("Populating %s json file.", args.file)
        json_file = args.file
        if exists_file(json_file):
            json_data = json_from_file(json_file)
            if json_data and 'fileType' in json_data:
                filetype = json_data['fileType']
            elif args.type:
                filetype = args.type
            else:
                filetype = 'structure'
            logger.info('Storing file:%s', json_file)
            db_record = json_record(args.container, filetype, json_file, json_data)
            if validate_json_data(db_record['json'], db_record['type']):
                insert_one_document(db_record, db_record['collection'], dbname, False)
                logger.debug('DB Record: %s', json.dumps(db_record, indent=2))
            else:
                logger.info('Invalid json for type:%s', db_record['type'])
            logger.info('*' * 80)


def validate_json_data(json_data, filetype):
    try:
        valid = json_data['fileType'] == filetype
        if filetype == 'snapshot':
            valid = json_data['snapshots'] and isinstance(json_data['snapshots'], list)
        elif filetype == 'masterSnapshot':
            valid = json_data['snapshots'] and isinstance(json_data['snapshots'], list)
        elif filetype == 'test':
            valid = json_data['snapshot'] and json_data['testSet'] and \
                    isinstance(json_data['testSet'], list)
        elif filetype == 'mastertest':
            valid = json_data['masterSnapshot'] and json_data['testSet'] and \
                    isinstance(json_data['testSet'], list)
        elif filetype == 'structure':
            valid = True if json_data else False
        elif filetype == 'notifications':
            valid = True if json_data else False
    except Exception as ex:
        logger.info('Exception json population: %s', ex)
        valid = False if filetype != 'structure' else True
    return valid


def populate_json_main(arg_vals=None):
    """Main driver utility for running validator tests."""
    logger.info("Comand: '%s %s'", sys.executable.rsplit('/', 1)[-1], ' '.join(sys.argv))
    cmd_parser = argparse.ArgumentParser("Populate json files")
    cmd_parser.add_argument('container', action='store',
                            help='Container name for the json files.')
    cmd_parser.add_argument('--dir', action='store', default=None,
                            help='Populate all json files from this directory.')
    cmd_parser.add_argument('--file', action='store', default=None,
                            help='Populate only this file')
    cmd_parser.add_argument('--type', action='store', default='structure',
                            choices=['test', 'structure', 'snapshot', 'output', 'notifications'])

    args = cmd_parser.parse_args(arg_vals)
    # Delete the rundata at the end of the script.
    atexit.register(delete_currentdata)
    logger.info(args)
    init_currentdata()
    dbname, db_init_res = init_db()
    if db_init_res:
        for _, collection in collectiontypes.items():
            create_indexes(config_value(DATABASE, collection), dbname, [('timestamp', TEXT)])
        populate_json_files(args)
    else:
        logger.error("Error initializing DB, exiting....!")
    return 0
